from abc import ABCMeta, abstractmethod
from collections import Sequence as SequenceCollection
from multiprocessing import Process
from pathlib import Path
import sys
from typing import (
    Dict,
    Iterable,
    Optional,
    Sequence as SequenceType,
    Tuple,
    Type,
    Union,
)

from loguru import logger
from xphyle import STDOUT, STDERR

from atropos.adapters import AdapterParser, AdapterType
from atropos.commands import (
    BaseCommand,
    Summary,
    Pipeline,
    SingleEndPipelineMixin,
    PairedEndPipelineMixin,
)
from atropos.commands.multicore import WorkerProcess
from atropos.commands.metrics import (
    MetricsMode,
    SingleEndReadMetrics,
    PairedEndReadMetrics,
)
from atropos.commands.trim.modifiers import (
    AdapterCutter,
    DoubleEncoder,
    InsertAdapterCutter,
    LengthTagModifier,
    MergeOverlapping,
    MinCutter,
    NEndTrimmer,
    TwoColorQualityTrimmer,
    UmiTrimmer,
    SyncUmi,
    AddUmi,
    NonDirectionalBisulfiteTrimmer,
    OverwriteRead,
    PairedEndModifiers,
    PrefixSuffixAdder,
    PrimerTrimmer,
    QualityTrimmer,
    RRBSTrimmer,
    SingleEndModifiers,
    SuffixRemover,
    SwiftBisulfiteTrimmer,
    UnconditionalCutter,
    ZeroCapper,
    Modifiers,
)
from atropos.commands.trim.filters import (
    Filter,
    FilterFactory,
    Filters,
    MergedReadFilter,
    NContentFilter,
    NoFilter,
    TooLongReadFilter,
    TooShortReadFilter,
    TrimmedFilter,
    UntrimmedFilter,
)
from atropos.commands.trim.writers import (
    Formatters,
    InfoFormatter,
    RestFormatter,
    WildcardFormatter,
    Writers,
)
from atropos.io.sequence import Sequence
from atropos.utils import ReturnCode, classproperty, run_interruptible
from atropos.utils.argparse import Namespace
from atropos.utils.collections import Summarizable
from atropos.utils.statistics import RandomMatchProbability


class RecordHandler(Summarizable):
    """
    Base class for record handlers.
    """

    def __init__(self, modifiers: Modifiers, filters: Filters, formatters: Formatters):
        self.modifiers = modifiers
        self.filters = filters
        self.formatters = formatters

    def handle_record(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ) -> Tuple[Type[Filter], Tuple[Sequence, Optional[Sequence]]]:
        """
        Handles a pair of reads.
        """
        reads = self.modifiers.modify(read1, read2)
        dest = self.filters.filter(*reads)
        self.formatters.format(context["results"], dest, *reads)
        return dest, reads

    def summarize(self) -> dict:
        """
        Returns a summary dict.
        """
        return dict(
            trim=dict(
                modifiers=self.modifiers.summarize(),
                filters=self.filters.summarize(),
                formatters=self.formatters.summarize(),
            )
        )


class MetricsRecordHandlerWrapper(Summarizable):
    """
    Wrapper around a record handler that collects read metrics before and/or after
    trimming.
    """

    def __init__(
        self,
        record_handler: RecordHandler,
        paired: bool,
        metrics_args: Dict[MetricsMode, dict],
        **kwargs,
    ):
        """
        Args:
            record_handler:
            paired: Whether reads are paired-end.
            metrics_args: Sequence; when to collect metrics ('pre', 'post')
            kwargs:
                mode: Collection mode; pre=only collect pre-trim metrics; post=only
                collect post-trim metrics; both=collect both pre- and post-trim metrics.
        """
        self.record_handler = record_handler
        self.read_metrics_class = (
            PairedEndReadMetrics if paired else SingleEndReadMetrics
        )

        if MetricsMode.PRE in metrics_args:
            self.pre = {}
            self.pre_kwargs = kwargs.copy()
            self.pre_kwargs.update(metrics_args[MetricsMode.PRE])
        else:
            self.pre = None

        if MetricsMode.POST in metrics_args:
            self.post = {}
            self.post_kwargs = kwargs.copy()
            self.post_kwargs.update(metrics_args[MetricsMode.POST])
        else:
            self.post = None

    def handle_record(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ) -> Tuple[Type[Filter], Tuple[Sequence, Optional[Sequence]]]:
        """
        Handles a pair of reads.

        Args:
            context:
            read1:
            read2:
        """
        if self.pre is not None:
            self.collect(self.pre, context["source"], read1, read2, **self.pre_kwargs)
        dest, reads = self.record_handler.handle_record(context, read1, read2)
        if self.post is not None:
            if dest not in self.post:
                self.post[dest] = {}
            self.collect(self.post[dest], context["source"], *reads, **self.post_kwargs)
        return dest, reads

    def collect(
        self,
        metrics: dict,
        source: Union[str, Tuple[str, str]],
        read1: Sequence,
        read2: Optional[Sequence] = None,
        **kwargs,
    ):
        """
        Collects metrics on a pair of reads.

        Args:
            metrics: The :class:`ReadStatistics` object.
            source: The source file(s).
            read1
            read2: The reads.
            kwargs:
        """
        if source not in metrics:
            metrics[source] = self.read_metrics_class(**kwargs)

        metrics[source].collect(read1, read2)

    def summarize(self) -> dict:
        """
        Returns a summary dict.
        """
        summary = self.record_handler.summarize()

        if self.pre is not None:
            summary["pre"] = dict(
                (source, metrics.summarize()) for source, metrics in self.pre.items()
            )

        if self.post is not None:
            summary["post"] = {}
            for dest, metrics_dict in self.post.items():
                summary["post"][dest.name] = dict(
                    (source, metrics.summarize())
                    for source, metrics in metrics_dict.items()
                )

        return summary


class ResultHandler(metaclass=ABCMeta):
    """
    Base class for result handlers.
    """

    def start(self, worker: Optional[Process] = None):
        """
        Starts the result handler.
        """

    def finish(self, total_batches: Optional[int] = None):
        """
        Finishes the result handler.

        Args:
            total_batches: Total number of batches processed.
        """

    @abstractmethod
    def write_result(self, batch_num: int, result: dict):
        """
        Writes a batch of results to output.

        Args:
            batch_num: The batch number.
            result: The result to write.
        """


class ResultHandlerWrapper(ResultHandler):
    """
    Wraps a ResultHandler.
    """

    def __init__(self, handler: ResultHandler):
        self.handler = handler

    def start(self, worker: Optional[Process] = None):
        self.handler.start(worker)

    def write_result(self, batch_num: int, result: dict):
        self.handler.write_result(batch_num, result)

    def finish(self, total_batches: Optional[int] = None):
        self.handler.finish(total_batches=total_batches)


class WorkerResultHandler(ResultHandlerWrapper):
    """
    Wraps a ResultHandler and compresses results prior to writing.
    """

    def write_result(self, batch_num: int, result: dict):
        """
        Given a dict mapping files to lists of strings, joins the strings and
        compresses them (if necessary) and then returns the properly formatted
        result dict.
        """
        self.handler.write_result(
            batch_num, dict(self.prepare_file(*item) for item in result.items())
        )

    def prepare_file(self, path: str, strings: Iterable[str]):
        """
        Prepares data for writing.

        Returns:
            Tuple (path, data).
        """
        return path, "".join(strings)


class WriterResultHandler(ResultHandler):
    """
    ResultHandler that writes results to disk.
    """

    def __init__(
        self, writers: Writers, compressed: bool = False, use_suffix: bool = False
    ):
        """
        Args:
            writers: :class:`Writers` object.
            compressed: Whether the data is compressed.
            use_suffix: Whether to add the worker index as a file suffix. Used for
                parallel-write mode.
        """
        self.writers = writers
        self.compressed = compressed
        self.use_suffix = use_suffix

    def start(self, worker: Optional[WorkerProcess] = None):
        if self.use_suffix:
            if worker is None:
                raise ValueError("worker must not be None")

            self.writers.suffix = f".{worker.index}"

    def write_result(self, batch_num: int, result: dict):
        self.writers.write_result(result, self.compressed)

    def finish(self, total_batches: Optional[int] = None):
        self.writers.close()


class TrimPipeline(Pipeline, metaclass=ABCMeta):
    """
    Base trimming pipeline.

    Args:
        record_handler:
        result_handler:
    """

    def __init__(self, record_handler: RecordHandler, result_handler: ResultHandler):
        super().__init__()
        self.record_handler = record_handler
        self.result_handler = result_handler

    def start(self, worker: Optional[Process] = None):
        self.result_handler.start(worker)

    def add_to_context(self, context: dict):
        context["results"] = {}

    def handle_records(self, context: dict, records: SequenceType[Sequence]):
        super().handle_records(context, records)
        self.result_handler.write_result(context["index"], context["results"])

    def handle_reads(
        self, context: dict, read1: Sequence, read2: Optional[Sequence] = None
    ):
        return self.record_handler.handle_record(context, read1, read2)

    def finish(self, summary: Summary, **kwargs):
        self.result_handler.finish()
        super().finish(summary)
        summary.update(self.record_handler.summarize())


class TrimSummary(Summary):
    """
    Summary that adds aggregate values for record and bp metrics.
    """

    def _post_process_other(self, dict_val: dict, key, value) -> None:
        """
        For trim metrics, any value with a name that starts with 'records_' will have
        'fraction_<var>' computed as value / total_records, and any value with a name
        that starts with 'bp_' will have 'fraction_<var>' and 'total_<var>' computed.
        We also replace any `Const`s with their values.
        """
        if self.has_exception:
            return

        def frac(val: int, _total: int):
            """
            Computes fraction of total.
            """
            return (val / _total) if val and _total != 0 else 0

        if isinstance(key, str):
            if key.startswith("records_"):
                frac_key = f"fraction_{key}"
                total_records = self["total_record_count"]

                if isinstance(value, SequenceCollection):
                    dict_val[frac_key] = [frac(val, total_records) for val in value]
                    total = sum(val for val in value if val)
                    dict_val[f"total_{key}"] = total
                else:
                    dict_val[frac_key] = frac(value, total_records)
            elif key.startswith("bp_"):
                frac_key = f"fraction_{key}"
                sum_total_bp = self["sum_total_bp_count"]

                if isinstance(value, SequenceCollection):
                    dict_val[frac_key] = [
                        frac(val, bps)
                        for val, bps in zip(value, self["total_bp_counts"])
                    ]
                    total = sum(val for val in value if val)
                    dict_val[f"total_{key}"] = total
                    dict_val[f"fraction_total_{key}"] = frac(
                        total, sum_total_bp
                    )
                else:
                    dict_val[frac_key] = frac(value, sum_total_bp)


class TrimCommand(BaseCommand):
    @classproperty
    def name(cls) -> str:
        return "trim"

    @classmethod
    def _create_summary(cls) -> dict:
        return TrimSummary()

    def __call__(self) -> ReturnCode:
        options = self.options
        match_probability = RandomMatchProbability()

        # Create Adapters
        adapters1 = []
        adapters2 = []
        has_adapters1 = bool(options.adapters or options.anywhere or options.front)
        has_adapters2 = bool(options.adapters2 or options.anywhere2 or options.front2)

        if has_adapters1 or has_adapters2:
            adapter_cache = self._load_known_adapters()

            parser_args = dict(
                colorspace=options.colorspace,
                max_error_rate=options.error_rate,
                min_overlap=options.overlap,
                read_wildcards=options.match_read_wildcards,
                adapter_wildcards=options.match_adapter_wildcards,
                indels=options.indels,
                indel_cost=options.indel_cost,
                cache=adapter_cache,
                gc_content=options.gc_content,
                match_probability=match_probability,
                alphabet=options.alphabet,
            )
            if options.adapter_max_rmp:
                parser_args["max_rmp"] = options.adapter_max_rmp

            adapter_parser = AdapterParser(**parser_args)

            if has_adapters1:
                adapters1 = adapter_parser.parse_multi(
                    options.adapters, options.anywhere, options.front
                )

            if has_adapters2:
                adapters2 = adapter_parser.parse_multi(
                    options.adapters2, options.anywhere2, options.front2
                )

            if options.cache_adapters:
                adapter_cache.save()

        # Create Modifiers
        # TODO: can this be replaced with an argparse required group?
        if (
            not adapters1
            and not adapters2
            and not options.quality_cutoff
            and options.twocolor_trim is None
            and options.cut == []
            and options.cut2 == []
            and options.cut_min == []
            and options.cut_min2 == []
            and (options.minimum_length is None or options.minimum_length <= 0)
            and options.maximum_length == sys.maxsize
            and not options.trim_n
            and not self.get_option("has_qualfile")
            and options.max_n is None
            and (not options.paired or options.overwrite_low_quality is None)
        ):
            raise ValueError("You need to provide at least one adapter sequence.")

        if options.aligner == "insert" and any(
            not a or len(a) != 1 or a[0].where != AdapterType.BACK
            for a in (adapters1, adapters2)
        ):
            raise ValueError(
                "Insert aligner requires a single 3' adapter for each read"
            )

        if options.debug:
            for adapter in adapters1 + adapters2:
                adapter.enable_debug()

        if options.paired:
            modifiers = PairedEndModifiers(options.paired)
        else:
            modifiers = SingleEndModifiers()

        if options.read1_umi or options.read2_umi:
            modifiers.add_modifier_pair(
                UmiTrimmer,
                dict(number_of_bases=options.read1_umi),
                dict(number_of_bases=options.read2_umi),
            )
            if options.paired:
                modifiers.add_modifier(SyncUmi, delim=options.umi_delim)
            else:
                modifiers.add_modifier(AddUmi, delim=options.umi_delim)

        for oper in options.op_order:
            if oper == "W" and options.overwrite_low_quality:
                lowq, highq, window = options.overwrite_low_quality
                modifiers.add_modifier(
                    OverwriteRead,
                    worse_read_min_quality=lowq,
                    better_read_min_quality=highq,
                    window_size=window,
                    base=options.quality_base,
                )
            elif oper == "A" and (adapters1 or adapters2):
                # TODO: generalize this using some kind of factory class
                if options.aligner == "insert":
                    # Use different base probabilities if we're trimming
                    # bisulfite data.
                    # TODO: this doesn't seem to help things, so commenting it
                    #  out for now
                    # if options.bisulfite:
                    #   base_probs = dict(match_prob=0.33, mismatch_prob=0.67)
                    # else:
                    #   base_probs = dict(match_prob=0.25, mismatch_prob=0.75)
                    modifiers.add_modifier(
                        InsertAdapterCutter,
                        adapter1=adapters1[0],
                        adapter2=adapters2[0],
                        action=options.action,
                        mismatch_action=options.correct_mismatches,
                        max_insert_mismatch_frac=options.insert_match_error_rate,
                        max_adapter_mismatch_frac=
                        options.insert_match_adapter_error_rate,
                        match_probability=match_probability,
                        insert_max_rmp=options.insert_max_rmp,
                        read_wildcards=options.match_read_wildcards,
                        adapter_wildcards=options.match_adapter_wildcards,
                    )
                else:
                    a1_args = (
                        dict(
                            adapters=adapters1,
                            times=options.times,
                            action=options.action,
                        )
                        if adapters1
                        else None
                    )
                    a2_args = (
                        dict(
                            adapters=adapters2,
                            times=options.times,
                            action=options.action,
                        )
                        if adapters2
                        else None
                    )
                    modifiers.add_modifier_pair(AdapterCutter, a1_args, a2_args)
            elif oper == "C" and (options.cut or options.cut2):
                modifiers.add_modifier_pair(
                    UnconditionalCutter,
                    dict(lengths=options.cut),
                    dict(lengths=options.cut2),
                )
            elif oper == "G" and (options.twocolor_trim is not None):
                modifiers.add_modifier(
                    TwoColorQualityTrimmer,
                    cutoff=options.twocolor_trim,
                    base=options.quality_base,
                )
            elif oper == "Q" and options.quality_cutoff:
                modifiers.add_modifier(
                    QualityTrimmer,
                    cutoff_front=options.quality_cutoff[0],
                    cutoff_back=options.quality_cutoff[1],
                    base=options.quality_base,
                )

        if options.bisulfite:
            if isinstance(options.bisulfite, str):
                if "non-directional" in options.bisulfite:
                    modifiers.add_modifier(
                        NonDirectionalBisulfiteTrimmer,
                        rrbs=options.bisulfite == "non-directional-rrbs",
                    )
                elif options.bisulfite == "rrbs":
                    modifiers.add_modifier(RRBSTrimmer)
                elif options.bisulfite in ("epignome", "truseq"):
                    # Trimming leads to worse results
                    # modifiers.add_modifier(TruSeqBisulfiteTrimmer)
                    pass
                elif options.bisulfite == "swift":
                    modifiers.add_modifier(SwiftBisulfiteTrimmer)
            else:
                if options.bisulfite[0]:
                    modifiers.add_modifier(MinCutter, read=1, **(options.bisulfite[0]))
                if len(options.bisulfite) > 1 and options.bisulfite[1]:
                    modifiers.add_modifier(MinCutter, read=2, **(options.bisulfite[1]))

        if options.trim_n:
            modifiers.add_modifier(NEndTrimmer)

        if options.cut_min or options.cut_min2:
            modifiers.add_modifier_pair(
                MinCutter, dict(lengths=options.cut_min), dict(lengths=options.cut_min2)
            )

        if options.length_tag:
            modifiers.add_modifier(LengthTagModifier, length_tag=options.length_tag)

        if options.strip_suffix:
            modifiers.add_modifier(SuffixRemover, suffixes=options.strip_suffix)

        if options.prefix or options.suffix:
            modifiers.add_modifier(
                PrefixSuffixAdder, prefix=options.prefix, suffix=options.suffix
            )

        if options.double_encode:
            modifiers.add_modifier(DoubleEncoder)

        delivers_qualities = self.get_option("delivers_qualities")

        if options.zero_cap and delivers_qualities:
            modifiers.add_modifier(ZeroCapper, quality_base=options.quality_base)

        if options.trim_primer:
            modifiers.add_modifier(PrimerTrimmer)

        if options.merge_overlapping:
            modifiers.add_modifier(
                MergeOverlapping,
                min_overlap=options.merge_min_overlap,
                error_rate=options.merge_error_rate,
                mismatch_action=options.correct_mismatches,
            )

        # Create Formatters

        interleaved = options.use_interleaved_output

        if interleaved:
            output1 = options.interleaved_output
            output2 = None
        else:
            output1 = options.output
            output2 = options.paired_output

        if options.output_format is None:
            if delivers_qualities:
                options.output_format = "fastq"
            else:
                options.output_format = "fasta"

        formatters = Formatters(
            output1,
            dict(
                file_format=options.output_format,
                qualities=delivers_qualities,
                colorspace=options.colorspace,
                interleaved=interleaved,
            ),
        )

        # Create filters
        min_affected = 2 if options.pair_filter == "both" else 1
        filters = Filters(FilterFactory(options.paired, min_affected))
        force_create = []

        if options.merge_overlapping:
            filters.add_filter(MergedReadFilter)
            if options.merged_output:
                formatters.add_seq_formatter(MergedReadFilter, options.merged_output)

        if options.minimum_length is not None and options.minimum_length > 0:
            filters.add_filter(TooShortReadFilter, options.minimum_length)
            if options.too_short_output:
                formatters.add_seq_formatter(
                    TooShortReadFilter,
                    options.too_short_output,
                    options.too_short_paired_output,
                )

        if options.maximum_length < sys.maxsize:
            filters.add_filter(TooLongReadFilter, options.maximum_length)
            if options.too_long_output is not None:
                formatters.add_seq_formatter(
                    TooLongReadFilter,
                    options.too_long_output,
                    options.too_long_paired_output,
                )

        if options.max_n is not None:
            filters.add_filter(NContentFilter, options.max_n)

        if options.discard_trimmed:
            filters.add_filter(TrimmedFilter)

        if not formatters.multiplexed:
            if output1 is not None:
                formatters.add_seq_formatter(NoFilter, output1, output2)
                if output1 != STDOUT and options.writer_process:
                    force_create.append(output1)
                    if output2 is not None:
                        force_create.append(output2)
            elif not (options.discard_trimmed and options.untrimmed_output):
                formatters.add_seq_formatter(NoFilter, options.default_outfile)
                if (
                    options.default_outfile not in (STDOUT, STDERR) and
                    options.writer_process
                ):
                    force_create.append(options.default_outfile)

        if options.discard_untrimmed or options.untrimmed_output:
            filters.add_filter(UntrimmedFilter)

        if not options.discard_untrimmed:
            if formatters.multiplexed:
                untrimmed = (
                    options.untrimmed_output or
                    Path(str(output1).format(name="unknown"))
                )
                formatters.add_seq_formatter(UntrimmedFilter, untrimmed)
                formatters.add_seq_formatter(NoFilter, untrimmed)
            elif options.untrimmed_output:
                formatters.add_seq_formatter(
                    UntrimmedFilter,
                    options.untrimmed_output,
                    options.untrimmed_paired_output,
                )

        if options.rest_file:
            formatters.add_detail_formatter(RestFormatter(options.rest_file))

        if options.info_file:
            formatters.add_detail_formatter(InfoFormatter(options.info_file))

        if options.wildcard_file:
            formatters.add_detail_formatter(WildcardFormatter(options.wildcard_file))

        # Create writers
        writers = Writers(force_create, options.compression_format)

        # Create record handler
        record_handler = RecordHandler(modifiers, filters, formatters)

        if options.metrics:
            record_handler = MetricsRecordHandlerWrapper(
                record_handler,
                options.paired,
                options.metrics,
                qualities=delivers_qualities,
                quality_base=options.quality_base,
            )

        num_adapters = sum(len(a) for a in modifiers.get_adapters())
        trim_mode = {
            False: "single-end",
            "first": "paired-end legacy",
            "both": "paired-end",
        }[options.paired]
        logger.info(
            f"Trimming {num_adapters} adapter{'s' if num_adapters > 1 else ''} "
            f"with at most {options.error_rate * 100:.1f} errors in {trim_mode} mode..."
        )

        if options.paired == "first" and (
            len(record_handler.modifiers.get_modifiers(read=2)) > 0
            or options.quality_cutoff
        ):
            logger.warning(
                """Requested read modifications are applied only to the first read 
                since backwards compatibility mode is enabled. To modify both reads, 
                also use any of the -A/-B/-G/-U options. Use a dummy adapter sequence 
                when necessary: -A XXX"""
            )

        if options.paired:
            mixin_class = PairedEndPipelineMixin
        else:
            mixin_class = SingleEndPipelineMixin

        if options.threads is None:
            # Run single-threaded version
            result_handler = WorkerResultHandler(WriterResultHandler(writers))
            pipeline_class = type("TrimPipelineImpl", (mixin_class, TrimPipeline), {})
            pipeline = pipeline_class(record_handler, result_handler)
            self.summary.update(mode="serial", threads=1)
            return run_interruptible(pipeline, self, raise_on_error=True)
        else:
            # Run multiprocessing version
            self.summary.update(mode="parallel", threads=options.threads)
            return self.run_parallel(record_handler, writers, mixin_class)

    def run_parallel(
        self,
        record_handler: RecordHandler,
        writers: Writers,
        mixin_class: Type[Union[SingleEndPipelineMixin, PairedEndPipelineMixin]],
    ):
        """
        Parallel implementation of run_atropos. Works as follows:

        1. Main thread creates N worker processes (where N is the number of
        threads to be allocated) and (optionally) one writer process.
        2. Main thread loads batches of reads (or read pairs) from input file(s)
        and adds them to a queue (the input queue).
        3. Worker processes take batches from the input queue, process them as
        Atropos normally does, and either add the results to the result queue
        (if using a writer process) or write the results to disk. Each result is
        a dict mapping output file names to strings, where each string is a
        concatenation of reads (with appropriate line endings) to be written.
        A parameter also controls whether data compression is done by the workers or
        the writer.
        4. If using a writer process, it takes results from the result queue and
        writes each string to its corresponding file.
        5. When the main process finishes loading reads from the input file(s),
        it sends a signal to the worker processes that they should complete when
        the input queue is empty. It also singals the writer process how many
        total batches to expect, and the writer process exits after it has processed
        that many batches.
        6. When a worker process completes, it adds a summary of its activity to
        the summary queue.
        7. The main process reads summaries from the summary queue and merges
        them to create the complete summary, which is used to generate the report.

        There are several possible points of failure:

        1. The main process may exit due to an unexpected error, or becuase the
        user forces it to exit (Ctrl-C). In this case, an attempt is made to
        cancel all processes before exiting.
        2. A worker or writer process may exit due to an unknown error. To
        handle this, the main process checks that each process is alive whenver
        it times out writing to the input queue, and again when waiting for
        worker summaries. If a process has died, the program exits with an error
        since some data might have gotten lost.
        3. More commonly, process will time out blocking on reading from or
        writing to a queue. Size limits are used (optionally) for the input and
        result queues to prevent using lots of memory. When few threads are
        allocated, it is most likely that the main and writer processes will
        block, whereas with many threads allocated the workers are most likely
        to block. Also, e.g. in a cluster environment, I/O latency may cause a
        "backup" resulting in frequent blocking of the main and workers
        processes. Finally, also e.g. in a cluster environment, processes may
        suspended for periods of time. Use of a hard timeout period, after which
        processes are forced to exit, is thus undesirable. Instead, parameters
        are provided for the user to tune the batch size and max queue sizes to
        their particular environment. Additionally, a "soft" timeout is used,
        after which log messages are escallated from DEBUG to ERROR level. The
        user can then make the decision of whether or not to kill the program.

        Args:
            record_handler: RecordHandler object.
            writers: Writers object.
            mixin_class: Mixin to use for creating pipeline class.

        Returns:
            The return code.
        """
        # We do all the multicore imports and class definitions within the
        # run_parallel method to avoid extra overhead if only running in serial mode.
        from multiprocessing import Queue
        from atropos.commands.multicore import ParallelPipelineMixin
        from atropos.commands.trim.multicore import (
            CompressingWorkerResultHandler,
            Done,
            Killed,
            OrderPreservingWriterResultHandler,
            ParallelTrimPipelineRunner,
            QueueResultHandler,
            ResultProcess,
            WriterManager,
        )
        from atropos.utils.multicore import DEFAULT_RETRY_INTERVAL

        # Main process
        timeout = max(self.get_option("process_timeout"), DEFAULT_RETRY_INTERVAL)
        threads = self.get_option("threads")
        logger.debug(
            f"Starting atropos in parallel mode with threads={threads}, "
            f"timeout={timeout}"
        )

        if threads < 2:
            raise ValueError("'threads' must be >= 2")

        compression_mode = self.get_option("compression_mode")

        if compression_mode is None:
            compression_mode = choose_compression_mode(self.options)

        # Reserve a thread for the writer process if it will be doing the compression
        # and if one is available.
        if compression_mode == "writer" and threads > 2:
            threads -= 1

        # Queue by which results are sent from the worker processes to the writer
        # process
        result_queue = Queue(self.get_option("result_queue_size"))
        writer_manager = None

        if self.get_option("writer_process"):
            if compression_mode == "writer":
                worker_result_handler = WorkerResultHandler(
                    QueueResultHandler(result_queue)
                )
            else:
                worker_result_handler = CompressingWorkerResultHandler(
                    QueueResultHandler(result_queue),
                    compression_format=self.get_option("compression_format")
                )

            writer_manager = WriterManager(
                writers,
                compression_mode,
                self.get_option("preserve_order"),
                result_queue,
                timeout,
            )
        else:
            worker_result_handler = WorkerResultHandler(
                WriterResultHandler(writers, use_suffix=True)
            )

        pipeline_class = type(
            "TrimPipelineImpl", (ParallelPipelineMixin, mixin_class, TrimPipeline), {}
        )
        pipeline = pipeline_class(record_handler, worker_result_handler)

        runner = ParallelTrimPipelineRunner(self, pipeline, threads, writer_manager)

        return runner.run()


def choose_compression_mode(options: Namespace) -> str:
    if (
        options.output is None or
        options.output in (STDOUT, STDERR) or (
            options.writer_process and
            2 < options.threads < 8 and
            options.can_use_system_compression
        )
    ):
        # We must use a writer process to write to stdout/stderr. Otherwise, our
        # tests show that with 8 or more threads, worker compression is more efficient.
        return "writer"
    else:
        return "worker"



from collections import defaultdict
from io import StringIO
import os
import random
from textwrap import dedent

import pytest
from xphyle import open_, xopen

from atropos.errors import FormatError
from atropos.io import InputRead
from atropos.io.sequence import Sequence, ColorspaceSequence
from atropos.io.formatters import (
    FastaFormat,
    FastqFormat,
    InterleavedFormatter,
    SingleEndFormatter,
    SingleEndSAMFormatter,
    PairedEndSAMFormatter,
    create_seq_formatter,
)
from atropos.io.readers import (
    FastaReader,
    FastqReader,
    FastaQualReader,
    InterleavedSequenceReader,
    open_reader,
    sequence_names_match
)
from atropos.utils.ngs import ALPHABETS
from .utils import cutpath, datapath


# files tests/data/simple.fast{q,a}
simple_fastq = [
    Sequence("first_sequence", "SEQUENCE1", ":6;;8<=:<"),
    Sequence("second_sequence", "SEQUENCE2", "83<??:(61"),
]
simple_fasta = [Sequence(x.name, x.sequence, None) for x in simple_fastq]


class TestAlphabet:
    def test_alphabet(self):
        alphabet = ALPHABETS["dna"]
        for base in ("A", "C", "G", "T", "N"):
            assert base in alphabet
        assert "X" not in alphabet
        assert alphabet.resolve("X") == "N"


class TestSequence:
    def test_too_many_qualities(self):
        with pytest.raises(FormatError):
            Sequence(name="name", sequence="ACGT", qualities="#####")

    def test_too_many_qualities_colorspace(self):
        with pytest.raises(FormatError):
            ColorspaceSequence(name="name", sequence="T0123", qualities="#####")

    def test_invalid_primer(self):
        with pytest.raises(FormatError):
            ColorspaceSequence(name="name", sequence="K0123", qualities="####")


class TestFastaReader:
    def test(self):
        with pytest.raises(ValueError):
            with FastaReader(None) as _:
                pass

        with FastaReader(datapath("simple.fasta")) as f:
            reads = list(f)
        assert reads == simple_fasta

        fasta = StringIO(">first_sequence\nSEQUENCE1\n>second_sequence\nSEQUENCE2\n")
        reads = list(FastaReader(fasta))
        assert reads == simple_fasta

    def test_with_comments(self):
        fasta = StringIO(
            dedent(
                """
            # a comment
            # another one
            >first_sequence
            SEQUENCE1
            >second_sequence
            SEQUENCE2
            """
            )
        )
        reads = list(FastaReader(fasta))
        assert reads == simple_fasta

    def test_wrong_format(self):
        with pytest.raises(FormatError):
            fasta = StringIO(
                dedent(
                    """
                # a comment
                # another one
                unexpected
                >first_sequence
                SEQUENCE1
                >second_sequence
                SEQUENCE2
                """
                )
            )
            _ = list(FastaReader(fasta))

    def test_fastareader_keeplinebreaks(self):
        with FastaReader(datapath("simple.fasta"), keep_linebreaks=True) as f:
            reads = list(f)
        assert reads[0] == simple_fasta[0]
        assert reads[1].sequence == "SEQUEN\nCE2"

    def test_context_manager(self):
        filename = datapath("simple.fasta")
        with open(filename) as f:
            assert not f.closed
            _ = list(open_reader(f))
            assert not f.closed
        assert f.closed
        with FastaReader(filename) as sr:
            tmp_sr = sr
            assert not sr._file.closed
            _ = list(sr)
            assert not sr._file.closed
        assert tmp_sr._file is None
        # Open it a second time
        with FastaReader(filename) as _:
            pass


class TestFastqReader:
    def test_fastqreader(self):
        with FastqReader(datapath("simple.fastq")) as f:
            reads = list(f)
        assert reads == simple_fastq

    def test_fastqreader_dos(self):
        with FastqReader(datapath("dos.fastq")) as f:
            dos_reads = list(f)
        with FastqReader(datapath("small.fastq")) as f:
            unix_reads = list(f)
        assert dos_reads == unix_reads

    def test_fastq_wrongformat(self):
        with pytest.raises(FormatError), FastqReader(datapath("withplus.fastq")) as f:
            _ = list(f)

    def test_fastq_incomplete(self):
        fastq = StringIO("@name\nACGT+\n")
        with pytest.raises(FormatError), FastqReader(fastq) as fq:
            list(fq)

    def test_context_manager(self):
        filename = datapath("simple.fastq")
        with open(filename) as f:
            assert not f.closed
            _ = list(open_reader(f))
            assert not f.closed
        assert f.closed
        with FastqReader(filename) as sr:
            tmp_sr = sr
            assert not sr._file.closed
            _ = list(sr)
            assert not sr._file.closed
        assert tmp_sr._file is None

    def test_alphabet(self):
        filename = datapath("bad_bases.fq")
        with FastqReader(filename, alphabet=ALPHABETS["dna"]) as f:
            reads = list(f)
            assert reads[0].sequence == "ACGNGGACT"
            assert reads[1].sequence == "CGGACNNNC"


class TestFastaQualReader:
    def test_mismatching_read_names(self):
        with pytest.raises(FormatError):
            fasta = StringIO(">name\nACG")
            qual = StringIO(">nome\n3 5 7")
            list(FastaQualReader(fasta, qual))

    def test_invalid_quality_value(self):
        with pytest.raises(FormatError):
            fasta = StringIO(">name\nACG")
            qual = StringIO(">name\n3 xx 7")
            list(FastaQualReader(fasta, qual))


class TestSeqioOpen:
    def test_sequence_reader(self):
        # test the autodetection
        with open_reader(datapath("simple.fastq")) as f:
            reads = list(f)
        assert reads == simple_fastq
        with open_reader(datapath("simple.fasta")) as f:
            reads = list(f)
        assert reads == simple_fasta
        with open(datapath("simple.fastq")) as f:
            reads = list(open_reader(f))
        assert reads == simple_fastq
        # make the name attribute unavailable
        with open(datapath("simple.fastq")) as inp:
            f = StringIO(inp.read())
        reads = list(open_reader(f))
        assert reads == simple_fastq
        with open(datapath("simple.fasta")) as inp:
            f = StringIO(inp.read())
        reads = list(open_reader(f))
        assert reads == simple_fasta

    def test_autodetect_fasta_format(self, tmp_path):
        path = tmp_path / "tmp.fasta"
        fmt = create_seq_formatter(path)
        assert isinstance(fmt, SingleEndFormatter)
        assert isinstance(fmt._seq_format, FastaFormat)
        write_seq_output(simple_fasta, fmt)
        assert list(open_reader(path)) == simple_fasta

    def test_write_qualities_to_fasta(self, tmp_path):
        path = tmp_path / "tmp.fasta"
        fmt = create_seq_formatter(path, qualities=True)
        assert isinstance(fmt, SingleEndFormatter)
        assert isinstance(fmt._seq_format, FastaFormat)
        write_seq_output(simple_fasta, fmt)
        assert list(open_reader(path)) == simple_fasta

    def test_autodetect_fastq_format(self, tmp_path):
        path = tmp_path / "tmp.fastq"
        fmt = create_seq_formatter(path)
        assert isinstance(fmt, SingleEndFormatter)
        assert isinstance(fmt._seq_format, FastqFormat)
        write_seq_output(simple_fastq, fmt)
        assert list(open_reader(path)) == simple_fastq

    def test_fastq_qualities_missing(self, tmp_path):
        with pytest.raises(ValueError):
            path = tmp_path / "tmp.fastq"
            create_seq_formatter(path, qualities=False)


def write_seq_output(reads, fmt):
    result = defaultdict(list)
    for read in reads:
        fmt.format(result, read)
    for path, seqs in result.items():
        with xopen(path, "w") as f:
            f.write("".join(seqs))


class TestInterleavedReader:
    def test(self):
        expected = [
            (
                Sequence("read1/1 some text", "TTATTTGTCTCCAGC", "##HHHHHHHHHHHHH"),
                Sequence("read1/2 other text", "GCTGGAGACAAATAA", "HHHHHHHHHHHHHHH"),
            ),
            (
                Sequence("read3/1", "CCAACTTGATATTAATAACA", "HHHHHHHHHHHHHHHHHHHH"),
                Sequence("read3/2", "TGTTATTAATATCAAGTTGG", "#HHHHHHHHHHHHHHHHHHH"),
            ),
        ]
        reads = list(InterleavedSequenceReader(cutpath("interleaved.fastq")))
        assert reads == expected

        with open_reader(
            cutpath("interleaved.fastq"),
            interleaved=True,
            input_read=InputRead.PAIRED
        ) as f:
            reads = list(f)
        assert reads == expected

    def test_missing_partner(self):
        with pytest.raises(FormatError):
            s = StringIO("@r1\nACG\n+\nHHH")
            list(InterleavedSequenceReader(s))

    def test_incorrectly_paired(self):
        with pytest.raises(FormatError):
            s = StringIO("@r1/1\nACG\n+\nHHH\n@wrong_name\nTTT\n+\nHHH")
            list(InterleavedSequenceReader(s))


class TestFastaWriter:
    def test(self, tmp_path):
        path = tmp_path / "tmp"
        fmt = FastaFormat()
        with open_(path, "w") as fw:
            fw.write(fmt.format_entry("name", "CCATA"))
            fw.write(fmt.format_entry("name2", "HELLO"))
        with open_(path) as t:
            assert t.read() == ">name\nCCATA\n>name2\nHELLO\n"

    def test_linelength(self, tmp_path):
        path = tmp_path / "tmp"
        fmt = FastaFormat(line_length=3)
        with open_(path, "w") as fw:
            fw.write(fmt.format_entry("r1", "ACG"))
            fw.write(fmt.format_entry("r2", "CCAT"))
            fw.write(fmt.format_entry("r3", "TACCAG"))
        with open_(path) as t:
            x = t.read()
            print(x)
            assert x == ">r1\nACG\n>r2\nCCA\nT\n>r3\nTAC\nCAG\n"

    def test_write_sequence_object(self, tmp_path):
        path = tmp_path / "tmp"
        fmt = FastaFormat()
        with open_(path, "w") as fw:
            fw.write(fmt.format(Sequence("name", "CCATA")))
            fw.write(fmt.format(Sequence("name2", "HELLO")))
        with open_(path) as t:
            assert t.read() == ">name\nCCATA\n>name2\nHELLO\n"

    def test_write_zero_length_sequence(self):
        fmt = FastaFormat()
        s = fmt.format_entry("name", "")
        assert s == ">name\n\n", "{0!r}".format(s)


class TestFastqWriter:
    def test(self, tmp_path):
        path = tmp_path / "tmp.fastq"
        fmt = FastqFormat()
        with open_(path, "w") as fw:
            fw.write(fmt.format_entry("name", "CCATA", "!#!#!"))
            fw.write(fmt.format_entry("name2", "HELLO", "&&&!&&"))
        with open_(path) as t:
            assert t.read() == "@name\nCCATA\n+\n!#!#!\n@name2\nHELLO\n+\n&&&!&&\n"

    def test_twoheaders(self, tmp_path):
        path = tmp_path / "tmp.fastq"
        fmt = FastqFormat()
        with open_(path, "w") as fw:
            fw.write(fmt.format(Sequence("name", "CCATA", "!#!#!", name2="name")))
            fw.write(fmt.format(Sequence("name2", "HELLO", "&&&!&", name2="name2")))
        with open_(path) as t:
            assert (
                t.read() == "@name\nCCATA\n+name\n!#!#!\n@name2\nHELLO\n+name2\n&&&!&\n"
            )


class TestInterleavedWriter:
    def test(self):
        reads = [
            (
                Sequence("A/1 comment", "TTA", "##H"),
                Sequence("A/2 comment", "GCT", "HH#"),
            ),
            (Sequence("B/1", "CC", "HH"), Sequence("B/2", "TG", "#H")),
        ]
        fmt = InterleavedFormatter("foo", FastqFormat())
        result = defaultdict(lambda: [])
        for read1, read2 in reads:
            fmt.format(result, read1, read2)
        assert fmt.written == 2
        assert fmt.read1_bp == 5
        assert fmt.read2_bp == 5
        assert "foo" in result
        assert "".join(result["foo"]) == (
            "@A/1 comment\nTTA\n+\n##H\n@A/2 comment\nGCT\n+\nHH#\n@B/1\nCC\n+\nHH\n"
            "@B/2\nTG\n+\n#H\n"
        )


class TestSAMWriter:
    def test_single_end(self):
        reads = [Sequence("A/1", "TTA", "##H"), Sequence("B/1", "CC", "HH")]
        fmt = SingleEndSAMFormatter("foo")
        result = defaultdict(lambda: [])
        for read in reads:
            fmt.format(result, read)
        assert fmt.written == 2
        assert fmt.read1_bp == 5
        assert fmt.read2_bp == 0
        assert "foo" in result
        result_str = "".join(result["foo"])
        expected = (
            "@HD\tVN:1.6\tSO:unsorted\nA/1\t0\t*\t0\t0\t*\t*\t0\t0\tTTA\t##H\nB/1\t0"
            "\t*\t0\t0\t*\t*\t0\t0\tCC\tHH\n"
        )
        print(result_str)
        print(expected)
        assert result_str == expected

    def test_paired_end(self):
        reads = [
            (Sequence("A/1", "TTA", "##H"), Sequence("A/2", "GCT", "HH#")),
            (Sequence("B/1", "CC", "HH"), Sequence("B/2", "TG", "#H")),
        ]
        fmt = PairedEndSAMFormatter("foo")
        result = defaultdict(lambda: [])
        for read1, read2 in reads:
            fmt.format(result, read1, read2)
        assert fmt.written == 2
        assert fmt.read1_bp == 5
        assert fmt.read2_bp == 5
        assert "foo" in result
        result_str = "".join(result["foo"])
        expected = (
            "@HD\tVN:1.6\tSO:unsorted\nA/1\t65\t*\t0\t0\t*\t*\t0\t0\tTTA\t##H\nA/2\t129"
            "\t*\t0\t0\t*\t*\t0\t0\tGCT\tHH#\nB/1\t65\t*\t0\t0\t*\t*\t0\t0\tCC\tHH\nB/2"
            "\t129\t*\t0\t0\t*\t*\t0\t0\tTG\t#H\n"
        )
        assert result_str == expected


class TestPairedSequenceReader:
    def test_sequence_names_match(self):
        def match(name1, name2):
            seq1 = Sequence(name1, "ACGT")
            seq2 = Sequence(name2, "AACC")
            return sequence_names_match(seq1, seq2)

        assert match("abc", "abc")
        assert match("abc/1", "abc/2")
        assert match("abc.1", "abc.2")
        assert match("abc1", "abc2")
        assert not match("abc", "xyz")


try:
    import ngstream
    import ngs

    ngs_available = True
except ModuleNotFoundError:
    ngs_available = False


SRA_ACCESSION = "ERR2009169"
SRA_SEQ1 = Sequence(
    "D00442:178:C87AVANXX:2:2203:1465:2211",
    "CAGCTTCTTCATCATGTCCTCTACTTTCTTGGCCCGCTCGGCAGGCCCCAAC",
    "CCCCCFGGGGGGGGGGGGGGGFGGGGGGGGGCGGGGGGGG@BEDDGGGGGGG"
)


#@pytest.mark.skipif(not ngs_available, reason="ngstream library not available")
@pytest.mark.skip
class TestSraReader:
    def test_sra_reader(self):
        with ngstream.open(SRA_ACCESSION, "sra") as stream:
            reader = open_reader(ngstream_reader=stream)
            assert next(iter(reader)) == (SRA_SEQ1,)


def create_truncated_file(path):
    # Random text
    text = "".join(random.choice("ABCDEFGHIJKLMNOPQRSTUVWXYZ") for _ in range(200))
    f = xopen(path, "w")
    f.write(text)
    f.close()
    f = open(path, "a")
    f.truncate(os.stat(path).st_size - 10)
    f.close()


def test_truncated_gz(tmp_path):
    with pytest.raises(EOFError):
        path = tmp_path / "truncated.gz"
        create_truncated_file(path)
        f = xopen(path, "r")
        f.read()
        f.close()


def test_truncated_gz_iter(tmp_path):
    with pytest.raises(EOFError):
        path = tmp_path / "truncated.gz"
        create_truncated_file(path)
        f = xopen(path, "r", use_system=False)  # work around bug in py3.4
        for _ in f:
            pass
        f.close()



from pytest import raises

from atropos.adapters import (
    Adapter,
    AdapterMatch,
    AdapterType,
    ColorspaceAdapter,
    InsertAligner,
    LinkedAdapter,
    parse_braces,
)
from atropos.io.sequence import Sequence

from .utils import approx_equal


def test_issue_52():
    adapter = Adapter(
        sequence="GAACTCCAGTCACNNNNN",
        where=AdapterType.BACK,
        max_error_rate=0.12,
        min_overlap=5,
        read_wildcards=False,
        adapter_wildcards=True,
    )
    read = Sequence(name="abc", sequence="CCCCAGAACTACAGTCCCGGC")
    am = AdapterMatch(
        astart=0,
        astop=17,
        rstart=5,
        rstop=21,
        matches=15,
        errors=2,
        front=None,
        adapter=adapter,
        read=read,
    )
    assert am.wildcards() == "GGC"
    """
    The result above should actually be 'CGGC' since the correct
    alignment is this one:

    adapter         GAACTCCAGTCACNNNNN
    mismatches           X     X
    read       CCCCAGAACTACAGTC-CCGGC

    Since we do not keep the alignment, guessing 'GGC' is the best we
    can currently do.
    """


def test_issue_80():
    # This issue turned out to not be an actual issue with the alignment
    # algorithm. The following alignment is found because it has more matches
    # than the 'obvious' one:
    #
    # TCGTATGCCGTCTTC
    # =========X==XX=
    # TCGTATGCCCTC--C
    #
    # This is correct, albeit a little surprising, since an alignment without
    # indels would have only two errors.
    adapter = Adapter(
        sequence="TCGTATGCCGTCTTC",
        where=AdapterType.BACK,
        max_error_rate=0.2,
        min_overlap=3,
        read_wildcards=False,
        adapter_wildcards=False,
    )
    read = Sequence(name="seq2", sequence="TCGTATGCCCTCC")
    result = adapter.match_to(read)
    assert read.original_length == 13, result
    assert result.errors == 3, result
    assert result.astart == 0, result
    assert result.astop == 15, result


def test_str():
    a = Adapter("ACGT", where=AdapterType.BACK, max_error_rate=0.1)
    str(a)
    str(a.match_to(Sequence(name="seq", sequence="TTACGT")))
    ca = ColorspaceAdapter("0123", where=AdapterType.BACK, max_error_rate=0.1)
    str(ca)


def test_color():
    with raises(ValueError):
        ColorspaceAdapter("0123", where=AdapterType.FRONT, max_error_rate=0.1)


def test_parse_braces():
    assert parse_braces("") == ""
    assert parse_braces("A") == "A"
    assert parse_braces("A{0}") == ""
    assert parse_braces("A{1}") == "A"
    assert parse_braces("A{2}") == "AA"
    assert parse_braces("A{2}C") == "AAC"
    assert parse_braces("ACGTN{3}TGACCC") == "ACGTNNNTGACCC"
    assert parse_braces("ACGTN{10}TGACCC") == "ACGTNNNNNNNNNNTGACCC"
    assert parse_braces("ACGTN{3}TGA{4}CCC") == "ACGTNNNTGAAAACCC"
    assert parse_braces("ACGTN{0}TGA{4}CCC") == "ACGTTGAAAACCC"


def test_parse_braces_fail():
    for expression in [
        "{",
        "}",
        "{}",
        "{5",
        "{1}",
        "A{-7}",
        "A{",
        "A{1",
        "N{7",
        "AN{7",
        "A{4{}",
        "A{4}{3}",
        "A{b}",
        "A{6X}",
        "A{X6}",
    ]:
        with raises(ValueError):
            parse_braces(expression)


def test_linked_adapter():
    linked_adapter = LinkedAdapter("AAAA", "TTTT")
    sequence = Sequence(name="seq", sequence="AAAACCCCCTTTT")
    match = linked_adapter.match_to(sequence)
    trimmed = linked_adapter.trimmed(match)
    assert trimmed.name == "seq"
    assert trimmed.sequence == "CCCCC"


def test_random_match_probabilities():
    a = Adapter("A", AdapterType.BACK)
    rmp = a.random_match_probabilities()
    assert rmp == [1.0, 0.25]
    a = Adapter("AC", AdapterType.BACK, gc_content=0.4)
    rmp = a.random_match_probabilities()
    assert rmp == [1.0, 0.3, 0.06]


def test_match_probability():
    a = InsertAligner("TTAGACATAT", "CAGTGGAGTA")
    k = 3
    n = 5
    i3 = (120 / (6 * 2)) * (0.25 ** 3) * (0.75 ** 2)
    i4 = (120 / 24) * (0.25 ** 4) * 0.75
    i5 = 0.25 ** 5
    assert approx_equal(a.match_probability(k, n), i3 + i4 + i5, 0.0001)


def test_insert_align():
    a1_seq = "TTAGACATATGG"
    a2_seq = "CAGTGGAGTATA"
    aligner = InsertAligner(a1_seq, a2_seq)
    r1 = "AGTCGAGCCCATTGCAGACT" + a1_seq[0:10]
    r2 = "AGTCTGCAATGGGCTCGACT" + a2_seq[0:10]
    insert_match, match1, match2 = aligner.match_insert(r1, r2)
    assert match1.rstart == 20
    assert match1.length == 10
    assert match2.rstart == 20
    assert match2.length == 10


def test_short_adapter_overlap():
    a1_seq = "TTAGACATAT"
    a2_seq = "CAGTGGAGTA"
    seq1 = "GACAGGCCGTTTGAATGTTGACGGGATGTT"
    seq2 = "CATCCCGTCAACATTCAAACGGCCTGTCCA"
    aligner = InsertAligner(a1_seq, a2_seq)
    insert_match, match1, match2 = aligner.match_insert(seq1, seq2)
    assert match1.rstart == 28
    assert match1.length == 2
    assert match2.rstart == 28
    assert match2.length == 2


# TODO: implement this test
# def test_insert_align_different_lengths():
#     # Test for #51. This is the alignment:
#     # TTAGGCTATGGCTTCTCGGGTTGAGGCTACAAGTTTTGGACCCTCCAGAGCAAAGCAGGTCTCTTTGACATCAGCTGCACAGCACTTGTCTACAAAAGCTGCAAA
#     #                                                                                                                       AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT
#     # ATAGGCTATGGCTTCTCGAGTTGAAGCTACAAGTTTTGGACCCTCCAGAGCAAAGCAGGTCTCTTTGACATCAGCTGCACAGCACTTGTCTACAAAAGCTGCAAAAGATCGGAAGAGCGTCTCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGACGTATCATTAAAAAAAAAAACACATCACATCAACAAGATAACACGACTTCTCCATCCACAGTACCGATGACCTCAACATTAGT
#     #
#     # The inserts should align even though there is a gap at the 5' end of read1.
#     a1_seq = "AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG"
#     a2_seq = "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT"
#     seq1 = "TTTGCAGCTTTTGTAGACAAGTGCTGTGCAGCTGATGTCAAAGAGACCTGCTTTGCTCTGGAGGGTCCAAAACTTGTAGCCTCAACCCGAGAAGCCATAGCCTAA"
#     seq2 = "ATAGGCTATGGCTTCTCGAGTTGAAGCTACAAGTTTTGGACCCTCCAGAGCAAAGCAGGTCTCTTTGACATCAGCTGCACAGCACTTGTCTACAAAAGCTGCAAAAGATCGGAAGAGCGTCTCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGACGTATCATTAAAAAAAAAAACACATCACATCAACAAGATAACACGACTTCTCCATCCACAGTACCGATGACCTCAACATTAGT"



from abc import ABCMeta, abstractmethod
from importlib import import_module
from pathlib import Path
from typing import IO, Optional, Sequence

from xphyle import STDOUT, STDERR, open_

from atropos.io import InputRead
from atropos.utils import classproperty
from atropos.utils.argparse import Namespace


class ReportWriter(metaclass=ABCMeta):
    def __init__(self, name: str, options: Namespace):
        self.name = name
        self.options = options

    @property
    def extensions(self) -> Sequence[str]:
        return self.name,

    @abstractmethod
    def write_report(self, summary: dict, output_file: Path) -> None:
        pass

    def _get_output_file(
        self, output_file: Path, extensions: Optional[Sequence[str]] = None
    ) -> Path:
        if not extensions:
            extensions = self.extensions
        if output_file.suffix and output_file.suffix[1:] in extensions:
            return output_file
        else:
            return Path(str(output_file) + f".{extensions[0]}")


class BaseReportWriter(ReportWriter, metaclass=ABCMeta):
    def write_report(self, summary: dict, output_file: Path):
        with open_(self._get_output_file(output_file), "wt") as stream:
            self.serialize(summary, stream)

    @abstractmethod
    def serialize(self, summary: dict, stream: IO) -> None:
        pass


class TemplateReportWriter(BaseReportWriter):
    def __init__(
        self,
        name: str,
        options: Namespace,
        template_name: Optional[str] = None,
        template_paths: Sequence[Path] = None,
        template_globals: Optional[dict] = None
    ):
        """
        Args:
            name: Report format. Must match the file extension on a discoverable
                template.
            template_name: A template name to use, rather than auto-discover.
            template_paths: Sequence of paths to search for templates.
            template_globals: Dict of additional globals to add to the template
                environment.
        """
        super().__init__(name, options)
        self.template_name = template_name
        self.template_paths = template_paths
        self.template_globals = template_globals

    def serialize(self, summary: dict, stream: IO):
        """
        Generate a report using Jinja2.
        """
        import jinja2

        template_name = self.template_name or f"template.{self.name}"
        template_paths = self.template_paths or []

        try:
            env = jinja2.Environment(loader=jinja2.FileSystemLoader(template_paths))
            if self.template_globals:
                env.globals.update(self.template_globals)
            template = env.get_template(template_name)
        except:
            raise IOError(f"Could not load template file '{template_name}'")

        stream.write(template.render(summary=summary))


class DumpReportWriter(BaseReportWriter):
    def __init__(self, name: str, options: Namespace):
        super().__init__(name, options)

    def serialize(self, summary: dict, stream: IO):
        mod = import_module(self.name)
        mod.dump(summary, stream)


class BaseReportGenerator(metaclass=ABCMeta):
    """
    Base class for command report generators.
    """

    @classproperty
    def default_report_formats(cls) -> Sequence[str]:
        return ()

    @classmethod
    def generate_reports(cls, summary: dict, options: Namespace):
        """
        Generates report(s) from a summary.

        Args:
            summary: The summary dict.
            options:
        """
        cls._add_derived_data(summary)

        report_file = options.report_file
        report_formats = options.report_formats

        if not report_formats:
            if report_file not in (STDOUT, STDERR) and report_file.suffix:
                report_formats = (report_file.suffix[1:],)
            else:
                report_formats = cls.default_report_formats

        for fmt in report_formats:
            writer = cls._create_report_writer(fmt, options)
            writer.write_report(summary, report_file)

    @classmethod
    def _add_derived_data(cls, summary: dict) -> None:
        """
        Adds to the summary some additional fields that are useful in the report.
        """
        derived = {
            "mean_sequence_lengths": tuple(
                None if bp is None else bp / summary["total_record_count"]
                for bp in summary["total_bp_counts"]
            )
        }
        inp = summary["input"]
        fmt = inp["file_format"]
        if inp["input_read"] == InputRead.PAIRED:
            fmt += ", Paired"
        else:
            fmt += f", Read {inp['input_read']}"
        if inp["colorspace"]:
            fmt += ", Colorspace"
        if inp["interleaved"]:
            fmt += ", Interleaved"
        if inp["delivers_qualities"]:
            fmt += ", w/ Qualities"
        else:
            fmt += ", w/o Qualities"
        derived["input_format"] = fmt
        summary["derived"] = derived

    @classmethod
    def _create_report_writer(cls, fmt: str, options: Namespace) -> ReportWriter:
        if fmt in ("json", "yaml", "pickle"):
            return DumpReportWriter(fmt, options)
        else:
            try:
                return TemplateReportWriter(fmt, options)
            except IOError:
                raise ValueError(f"Unsupported format {fmt}")


def prettyprint_summary(summary: dict, outfile: Path = Path("summary.dump.txt")):
    """
    Pretty-prints the summary to a file. Mostly used for debugging.
    """
    from pprint import pprint

    with open_(outfile, "w") as out:
        pprint(summary, out)



from atropos.adapters import *
from atropos.commands.trim.modifiers import *
from atropos.io.sequence import Sequence
from atropos.utils.ngs import reverse_complement as rc

DUMMY_ADAPTER = Adapter("ACGT", AdapterType.FRONT)


def front_match(read):
    match = AdapterMatch(0, 2, 0, 2, 1, 0, True, adapter=DUMMY_ADAPTER, read=read)
    match_info = MatchInfo(
        "read",
        0,
        0,
        2,
        "",
        "AC",
        "ACGTAC",
        "adapter",
        "",
        "##",
        "######",
        True,
        2,
        2,
        2,
    )
    return match, [match_info]


def back_match(read):
    match = AdapterMatch(6, 8, 6, 8, 1, 0, False, adapter=DUMMY_ADAPTER, read=read)
    match_info = MatchInfo(
        "read",
        0,
        6,
        8,
        "ACGTAC",
        "GT",
        "",
        "adapter",
        "######",
        "##",
        "",
        False,
        2,
        2,
        2,
    )
    return match, [match_info]


def test_unconditional_cutter():
    s = Sequence("read1", "abcdefg")
    assert UnconditionalCutter(lengths=[2])(s).sequence == "cdefg"
    assert UnconditionalCutter(lengths=[-2])(s).sequence == "abcde"
    assert UnconditionalCutter(lengths=[100])(s).sequence == ""
    assert UnconditionalCutter(lengths=[-100])(s).sequence == ""


def test_nend_trimmer():
    trimmer = NEndTrimmer()
    seqs = ["NNNNAAACCTTGGNNN", "NNNNAAACNNNCTTGGNNN", "NNNNNN"]
    trims = ["AAACCTTGG", "AAACNNNCTTGG", ""]
    for seq, trimmed in zip(seqs, trims):
        _seq = Sequence("read1", seq, qualities="#" * len(seq))
        _trimmed = Sequence("read1", trimmed, qualities="#" * len(trimmed))
        assert trimmer(_seq) == _trimmed


def test_quality_trimmer():
    read = Sequence("read1", "ACGTTTACGTA", "##456789###")
    qt = QualityTrimmer(10, 10, 33)
    assert qt(read) == Sequence("read1", "GTTTAC", "456789")
    qt = QualityTrimmer(0, 10, 33)
    assert qt(read) == Sequence("read1", "ACGTTTAC", "##456789")
    qt = QualityTrimmer(10, 0, 33)
    assert qt(read) == Sequence("read1", "GTTTACGTA", "456789###")


def test_modifiers_single():
    m = SingleEndModifiers()
    m.add_modifier(UnconditionalCutter, lengths=[5])
    mod1 = m.get_modifiers(read=1)
    mod2 = m.get_modifiers(read=2)
    assert len(mod1) == 1
    assert isinstance(mod1[0], UnconditionalCutter)
    assert len(mod2) == 0
    # test single-end
    read = Sequence("read1", "ACGTTTACGTA", "##456789###")
    mod_read = m.modify(read)
    assert len(mod_read) == 1
    assert mod_read[0].sequence == "TACGTA"


def test_modifiers_paired_legacy():
    m = PairedEndModifiers(paired="first")
    m.add_modifier(UnconditionalCutter, lengths=[5])
    read1 = Sequence("read1", "ACGTTTACGTA", "##456789###")
    read2 = Sequence("read1", "ACGTTTACGTA", "##456789###")
    mod_read1, mod_read2 = m.modify(read1, read2)
    assert mod_read1.sequence == "TACGTA"
    assert mod_read2.sequence == "ACGTTTACGTA"


def test_modifiers_paired_both():
    m = PairedEndModifiers(paired="both")
    m.add_modifier(UnconditionalCutter, read=1 | 2, lengths=[5])
    mod1 = m.get_modifiers(read=1)
    mod2 = m.get_modifiers(read=2)
    assert len(mod1) == 1
    assert len(mod2) == 1
    assert isinstance(mod1[0], UnconditionalCutter)
    assert isinstance(mod2[0], UnconditionalCutter)
    read1 = Sequence("read1", "ACGTTTACGTA", "##456789###")
    read2 = Sequence("read1", "ACGTTTACGTA", "##456789###")
    mod_read1, mod_read2 = m.modify(read1, read2)
    assert mod_read1.sequence == "TACGTA"
    assert mod_read2.sequence == "TACGTA"


def test_min_cutter_t_t():
    unconditional_before = UnconditionalCutter((2, -2))
    unconditional_after = UnconditionalCutter((1, -1))
    min_trimmer = MinCutter((5, -5), True, True)
    read1 = Sequence("read1", "CAATCGATCGAACGTACCGAT")
    assert read1.clipped == [0, 0, 0, 0], str(read1.clipped)
    read1 = unconditional_before(read1)
    assert read1.sequence == "ATCGATCGAACGTACCG"
    assert read1.clipped == [2, 2, 0, 0], str(read1.clipped)
    # test without adapter trimming
    assert min_trimmer(read1).sequence == "ATCGATCGAACGTACCG"
    # test with adapter trimming
    read2 = read1[:]
    read2.sequence = "ATCGAACGTACCG"
    read2.match, read2.match_info = front_match(read2)
    read3 = min_trimmer(read2)
    assert read3.sequence == "TCGAACGTACCG", read3.sequence
    assert read3.clipped == [2, 2, 1, 0]
    # test with subsequent clipping
    read4 = unconditional_after(read2)
    assert read4.sequence == "TCGAACGTACC", read4.sequence
    assert read4.clipped == [2, 2, 1, 1], read4.clipped
    read5 = min_trimmer(read4)
    assert read5.sequence == "TCGAACGTACC", read5.sequence
    assert read5.clipped == [2, 2, 1, 1], read5.clipped


def test_min_cutter_f_t():
    unconditional_before = UnconditionalCutter((2, -2))
    unconditional_after = UnconditionalCutter((1, -1))
    min_trimmer = MinCutter((5, -5), False, True)
    read1 = Sequence("read1", "CAATCGATCGAACGTACCGAT")
    read1 = unconditional_before(read1)
    assert read1.sequence == "ATCGATCGAACGTACCG"
    assert read1.clipped == [2, 2, 0, 0]
    # test without adapter trimming
    assert min_trimmer(read1).sequence == "ATCGATCGAACGTACCG"
    # test with adapter trimming
    read2 = read1[:]
    read2.match, read2.match_info = front_match(read2)
    read2.sequence = "CGATCGAACGTACCG"
    read3 = min_trimmer(read2)
    assert read3.sequence == "GAACGTACCG", read3.sequence
    assert read3.clipped == [2, 2, 5, 0]
    # test with subsequent clipping
    read4 = unconditional_after(read2)
    assert read4.sequence == "GATCGAACGTACC"
    assert read4.clipped == [2, 2, 1, 1]
    read5 = min_trimmer(read4)
    assert read5.sequence == "GAACGTACC", read5.sequence
    assert read5.clipped == [2, 2, 5, 1]


def test_min_cutter_t_f():
    unconditional_before = UnconditionalCutter((2, -2))
    min_trimmer = MinCutter((4, -4), True, False)
    read1 = Sequence("read1", "CAATCGATCGAACGTACCGAT")
    read1 = unconditional_before(read1)
    assert read1.sequence == "ATCGATCGAACGTACCG"
    assert read1.clipped == [2, 2, 0, 0]
    # test without adapter trimming
    assert min_trimmer(read1).sequence == "CGATCGAACGTAC"


def test_non_directional_bisulfite_trimmer():
    trimmer = NonDirectionalBisulfiteTrimmer(rrbs=True)
    read2 = Sequence("read2", "CTATCGATC")
    read2.match, read2.match_info = back_match(read2)
    read3 = Sequence("read3", "CTATCGATCCA")
    # assert trimmer(read1).sequence == "ATCGATC"
    assert trimmer(read2).sequence == "CTATCGA"
    assert trimmer(read3).sequence == "CTATCGATCCA"


def test_truseq_trimmer():
    trimmer = TruSeqBisulfiteTrimmer()
    read1 = Sequence("read1", "CTATCGATCCACGAGACTAAC")
    assert trimmer(read1).sequence == "ATCCACGAGACTAAC"


def test_swift_trimmer():
    trimmer = SwiftBisulfiteTrimmer()
    seq = "".join(["ACGT"] * 30)
    read1 = Sequence("read1", seq)
    read2 = Sequence("read2", seq)
    trimmed = trimmer(read1, read2)
    assert trimmed[0].sequence == seq[:-10]
    assert trimmed[1].sequence == seq[10:]


def test_overlapping():
    trimmer = MergeOverlapping(min_overlap=10, error_rate=0.1)
    a1 = "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC"
    a2 = reverse_complement("AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA")
    frag = "CCAAGCAGACATTCACTCAGATTGCA"
    r1 = (frag + a1)[0:40]
    q1 = "#" * 40
    r2 = reverse_complement(a2 + frag)[0:40]
    q2 = "!" * 40
    parser = AdapterParser()
    adapter1 = parser.parse_from_spec(a1)
    adapter2 = parser.parse_from_spec(a2)
    cutter = AdapterCutter([adapter1, adapter2])
    read1 = Sequence("foo", r1, q1)
    read1 = cutter(read1)
    assert len(read1) == 26
    read2 = Sequence("foo", r2, q2)
    read2 = cutter(read2)
    assert len(read2) == 26
    # complete overlap
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert read1 == read1_merged
    # partial overlap
    read1.merged = False
    read2 = read2.subseq(0, 24)[2]
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert read1 == read1_merged
    # partial overlap r1, r2
    read1.merged = False
    read1 = read1.subseq(0, 24)[2]
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert len(read1_merged) == 26
    assert read1_merged.sequence == "CCAAGCAGACATTCACTCAGATTGCA"
    assert read1_merged.qualities == ("#" * 24) + ("!" * 2)
    # errors
    # round(0.1 * 24) = 2, so 2 errors should pass but 3 should not
    read1.merged = False
    r1_seq = list(read1.sequence)
    r1_seq[10] = reverse_complement(r1_seq[10])
    r1_seq[20] = reverse_complement(r1_seq[20])
    read1.sequence = "".join(r1_seq)
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert len(read1_merged) == 26
    assert read1_merged.sequence == "CCAAGCAGACTTTCACTCAGTTTGCA"
    assert read1_merged.qualities == ("#" * 24) + ("!" * 2)
    # too few overlapping bases
    read1.merged = False
    r1_seq[15] = reverse_complement(r1_seq[15])
    read1.sequence = "".join(r1_seq)
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged is False
    assert read2 is not None


def test_overlapping_with_error_correction():
    trimmer = MergeOverlapping(
        min_overlap=10, error_rate=0.1, mismatch_action=MismatchAction.LIBERAL
    )
    r1 = "AGATCGGAAGACCGTCATGTAGGGAAAGAGTGTAGATCTC"
    q1 = "FFFFFFFFFFF#FFFFFFFFFFFFFFFFFFFFF#######"
    r2 = reverse_complement("AGATCGGTAGAGCGTCGTGTAGGGAAATAGTGTAGATCTC")
    q2 = "".join(reversed("FFFFFFFFFFFFFFFF#FFFFFFFFFF#FFFFFFFFFFFF"))
    read1 = Sequence("foo", r1, q1)
    read2 = Sequence("foo", r2, q2)
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert read1_merged.sequence == "AGATCGGTAGAGCGTCATGTAGGGAAAGAGTGTAGATCTC"
    assert read1_merged.qualities == "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF#######"


def test_mismatched_adapter_overlaps():
    """
    This is a test case from real data. The adapter overlaps 1 less bp
    on the fw read than on the reverse read. We want to make sure that
    the extra 'A' base gets trimmed.
    adapter                                                                                                                                                          GATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG
    actual                                                               TTGTTTTTATGGAGAGAGTTTTAAGGTTTATTTTAGTTTTAAAGGATATTGTAGGTTAGAGGGAAAGTGTATGATGAAGGTATATATTGGTAGATCGGAAGAGCACACGTCTGAACTTCAGTCAC
    actual rc                          TATGTTCTTTCCCTTCACGTCTCTCTTCGGATCTTTATTGTGATGAGTTGAAAATAAAGGTTAAGTATAGATAAAAAAGTTATTATAGTTTAGAGGGTAAGTGTATGATGGAGTAAAATATTGGT
    adapter rc AATGATACGGCGACCACCGAGATCTACACTCTTTCCCTACACGACGCTCTTCCGATCT
    """
    r1 = "TTGTTTTTATGGAGAGAGTTTTAAGGTTTATTTTAGTTTTAAAGGATATTGTAGGTTAGAGGGAAAGTGTATGATGAAGGTATATATTGGTAGATCGGAAGAGCACACGTCTGAACTTCAGTCAC"
    r2 = "ACCAATATTTTACTCCATCATACACTTACCCTCTAAACTATAATAACTTTTTTATCTATACTTAACCTTTATTTTCAACTCATCACAATAAAGATCCGAAGAGAGACGTGAAGGGAAAGAACATA"
    # TruSeq index 7
    a1 = "GATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG"
    # TruSeq universal
    a2 = "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT"
    parser = AdapterParser()
    adapter1 = parser.parse_from_spec(a1)
    adapter2 = parser.parse_from_spec(a2)
    # the data has a fairly high error rate
    cutter = InsertAdapterCutter(
        adapter1, adapter2, max_insert_mismatch_frac=0.3, max_adapter_mismatch_frac=0.3
    )
    read1 = Sequence("foo", r1, "#" * 125)
    read2 = Sequence("foo", r2, "#" * 125)
    new_read1, new_read2 = cutter(read1, read2)
    assert (len(new_read1)) == 91
    assert (len(new_read2)) == 91
    assert (
        new_read1.sequence
        == "TTGTTTTTATGGAGAGAGTTTTAAGGTTTATTTTAGTTTTAAAGGATATTGTAGGTTAGAGGGAAAGTGTATGATGAAGGTATATATTGGT"
    )


def test_error_correction():
    a1 = "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC"
    a2 = "AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA"
    frag = "CCAAGCAGACATTCACTCAGATTGCA"
    correct_frag = "CCAAGTAGACATTCGCTCAGATTGCA"
    r1 = list(frag)
    # C>T at pos 6
    r1[5] = "T"
    q1 = ["#"] * 40
    # quality of read1 > quality of read2 at pos 6
    q1[5] = "A"
    r1 = ("".join(r1) + a1)[0:40]
    q1 = "".join(q1)
    r2 = list(frag)
    # A>G at pos 15
    r2[14] = "G"
    q2 = ["#"] * 40
    # quality of read2 > quality of read1 at pos 11
    q2[len(frag) - 15] = "A"
    r2 = reverse_complement(reverse_complement(a2) + "".join(r2))[0:40]
    q2 = "".join(q2)
    read1 = Sequence("foo", r1, q1)
    read2 = Sequence("foo", r2, q2)
    parser = AdapterParser()
    adapter1 = parser.parse_from_spec(a1)
    adapter2 = parser.parse_from_spec(a2)
    cutter = InsertAdapterCutter(
        adapter1, adapter2, mismatch_action=MismatchAction.LIBERAL
    )
    new_read1, new_read2 = cutter(read1, read2)
    assert len(new_read1) == 26
    assert new_read1.insert_overlap
    assert new_read1.sequence == correct_frag
    assert len(new_read2) == 26
    assert new_read2.insert_overlap
    assert new_read2.sequence == reverse_complement(correct_frag)


def test_error_correction_no_insert_match_one_adapter_match():
    a1 = "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC"
    a2 = "AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA"
    a2_mod = "ACATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA"
    frag = "CCAAGCAGACATTCACTCAGATTGCA"
    correct_frag = "CCAAGTAGACATTCGCTCAGATTGCA"
    r1 = list(frag)
    # C>T at pos 6
    r1[5] = "T"
    q1 = ["#"] * 40
    # quality of read1 > quality of read2 at pos 6
    q1[5] = "A"
    r1 = ("".join(r1) + a1)[0:40]
    q1 = "".join(q1)
    r2 = list(frag)
    # A>G at pos 15
    r2[14] = "G"
    q2 = ["#"] * 40
    # quality of read2 > quality of read1 at pos 11
    q2[len(frag) - 15] = "A"
    r2 = reverse_complement(reverse_complement(a2_mod) + "".join(r2))[0:40]
    q2 = "".join(q2)
    read1 = Sequence("foo", r1, q1)
    read2 = Sequence("foo", r2, q2)
    parser1 = AdapterParser()
    adapter1 = parser1.parse_from_spec(a1)
    # Allow zero mismatches to prevent adapter alignment
    parser2 = AdapterParser(max_error_rate=0)
    adapter2 = parser2.parse_from_spec(a2)
    # Allow zero mismatches to prevent insert alignment
    cutter = InsertAdapterCutter(
        adapter1,
        adapter2,
        mismatch_action=MismatchAction.LIBERAL,
        max_insert_mismatch_frac=0,
    )
    new_read1, new_read2 = cutter(read1, read2)
    assert len(new_read1) == 26
    assert not new_read1.insert_overlap
    assert new_read1.sequence == correct_frag
    assert len(new_read2) == 26
    assert not new_read2.insert_overlap
    assert new_read2.sequence == reverse_complement(correct_frag)


def test_error_correction_no_insert_match_two_adapter_matches():
    a1 = "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC"
    a2 = "AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA"
    frag = "CCAAGCAGACATTCACTCAGATTGCA"
    correct_frag = "CCAAGTAGACATTCGCTCAGATTGCA"
    r1 = list(frag)
    # C>T at pos 6
    r1[5] = "T"
    q1 = ["#"] * 40
    # quality of read1 > quality of read2 at pos 6
    q1[5] = "A"
    r1 = ("".join(r1) + a1)[0:40]
    q1 = "".join(q1)
    r2 = list(frag)
    # A>G at pos 15
    r2[14] = "G"
    q2 = ["#"] * 40
    # quality of read2 > quality of read1 at pos 11
    q2[len(frag) - 15] = "A"
    r2 = reverse_complement(reverse_complement(a2) + "".join(r2))[0:40]
    q2 = "".join(q2)
    read1 = Sequence("foo", r1, q1)
    read2 = Sequence("foo", r2, q2)
    parser = AdapterParser()
    adapter1 = parser.parse_from_spec(a1)
    adapter2 = parser.parse_from_spec(a2)
    # Allow zero mismatches to prevent insert alignment
    cutter = InsertAdapterCutter(
        adapter1,
        adapter2,
        mismatch_action=MismatchAction.LIBERAL,
        max_insert_mismatch_frac=0,
    )
    new_read1, new_read2 = cutter(read1, read2)
    assert len(new_read1) == 26
    assert not new_read1.insert_overlap
    assert new_read1.sequence == correct_frag
    assert len(new_read2) == 26
    assert not new_read2.insert_overlap
    assert new_read2.sequence == reverse_complement(correct_frag)


def test_error_correction_unequal_read_lengths():
    # Test case for issue #51
    read1 = Sequence(
        "read1",
        "TTTGCAGCTTTTGTAGACAAGTGCTGTGCAGCTGATGTCAAAGAGACCTGCTTTGCTCTGGAGGGTCCAAAACTTGTAGCCTCAACCCGAGAAGCCATAGCCTAA",
        "CCCCCFCGGGGGBFFAFC<?BEADCCF<FFFFGFFDFDFFGGGGCFGGC?DFFFEC;,===??DG==DDDFFFFG8DDD7+5;;DF*=)))10885D**58>6=0",
    )
    read2 = Sequence(
        "read1",
        "ATAGGCTATGGCTTCTCGAGTTGAAGCTACAAGTTTTGGACCCTCCAGAGCAAAGCAGGTCTCTTTGACATCAGCTGCACAGCACTTGTCTACAAAAGCTGCAAAAGATCGGAAGAGCGTCTCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGACGTATCATTAAAAAAAAAAACACATCACATCAACAAGATAACACGACTTCTCCATCCACAGTACCGATGACCTCAACATTAGT",
        "CCCCCG@FCFGGCFGGGGFEFGFGGFCFGGGFGFGGGGGGGGGGGGGGGGGGGGGGGGGGG9FGGGGGGGFGDFFGGGGGGGGGGGGGGGGG8;>@?@FEGGGGGGGGGGGGGGGGGGGGG=DDFAEFFFGF>B>EA):DFFBDFFB6CDEDDD9=99DD>55)580:A5)*)*;DD>**51:0118):)4))1***0:*)*)((***0*.(((((*)/.)1/(6((()1.)(((6).-----8<:C<73",
    )
    aligner = InsertAligner(
        "AGATCGGAAGAGCACACGTCTGAACTCCAGTCACACAGTGATCTCGTATGCCGTCTTCTGCTTG",
        "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT",
    )
    insert_match, adapter_match1, adapter_match2 = aligner.match_insert(
        read1.sequence, read2.sequence
    )
    ec = ErrorCorrectorMixin(MismatchAction.N)
    ec.correct_errors(read1, read2, insert_match, truncate_seqs=True)
    assert read1.corrected == 3
    assert read2.corrected == 3
    for i in (80, 86, 104):
        assert read1.sequence[i] == "N", "Read 1 not corrected to N at {}".format(i)
        assert read2.sequence[104 - i] == "N", "Read 2 not corrected to N at {}".format(
            104 - i
        )


def ints2quals(ints):
    return "".join(chr(i + 33) for i in ints)


def test_overwrite_read():
    overwrite = OverwriteRead(20, 40, 10)
    lowseq = "ACGT" * 5
    highseq = "TCAG" * 5
    # mean lowq > 20, mean highq > 40
    lowq = (11, 31, 16, 24, 16, 20, 17, 19, 21, 28) * 2
    highq = (22, 62, 32, 48, 32, 40, 34, 38, 42, 56) * 2
    read1 = Sequence("foo", lowseq, ints2quals(lowq))
    read2 = Sequence("foo", highseq, ints2quals(highq))
    new_read1, new_read2 = overwrite(read1, read2)
    assert new_read1.sequence == lowseq
    assert new_read1.qualities == ints2quals(lowq)
    assert new_read2.sequence == highseq
    assert new_read2.qualities == ints2quals(highq)
    assert new_read1.corrected == new_read2.corrected == 0
    # mean lowq < 20, mean highq > 40
    lowq = tuple(i - 1 for i in lowq)
    read1 = Sequence("foo", lowseq, ints2quals(lowq))
    new_read1, new_read2 = overwrite(read1, read2)
    assert new_read1.sequence == rc(highseq)
    assert new_read1.qualities == ints2quals(reversed(highq))
    assert new_read2.sequence == highseq
    assert new_read2.qualities == ints2quals(highq)
    assert new_read1.corrected == new_read2.corrected == 1
    # mean lowq < 20, mean highq < 40
    highq = tuple(i - 1 for i in highq)
    read2 = Sequence("foo", highseq, ints2quals(highq))
    new_read1, new_read2 = overwrite(read1, read2)
    assert new_read1.sequence == lowseq
    assert new_read1.qualities == ints2quals(lowq)
    assert new_read2.sequence == highseq
    assert new_read2.qualities == ints2quals(highq)
    assert new_read1.corrected == new_read2.corrected == 0


def test_umi_trim_pair():
    seq1 = "TTTGCAGCTTTTGTAGACAAGTGCTGTGCAGCTGATGTCAAAGAGACCTGCTTTGCTCTGGAGGGTCCAAAACTTGTAGCCTCAACCCGAGAAGCCATAGCCTAA"
    qual1 = "CCCCCFCGGGGGBFFAFC<?BEADCCF<FFFFGFFDFDFFGGGGCFGGC?DFFFEC;,===??DG==DDDFFFFG8DDD7+5;;DF*=)))10885D**58>6=0"
    seq2 = "ATAGGCTATGGCTTCTCGAGTTGAAGCTACAAGTTTTGGACCCTCCAGAGCAAAGCAGGTCTCTTTGACATCAGCTGCACAGCACTTGTCTACAAAAGCTGCAAAAGATCGGAAGAGCGTCTCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGACGTATCATTAAAAAAAAAAACACATCACATCAACAAGATAACACGACTTCTCCATCCACAGTACCGATGACCTCAACATTAGT"
    qual2 = "CCCCCG@FCFGGCFGGGGFEFGFGGFCFGGGFGFGGGGGGGGGGGGGGGGGGGGGGGGGGG9FGGGGGGGFGDFFGGGGGGGGGGGGGGGGG8;>@?@FEGGGGGGGGGGGGGGGGGGGGG=DDFAEFFFGF>B>EA):DFFBDFFB6CDEDDD9=99DD>55)580:A5)*)*;DD>**51:0118):)4))1***0:*)*)((***0*.(((((*)/.)1/(6((()1.)(((6).-----8<:C<73"

    # 5 bases UMI from both reads
    read1 = Sequence("read1", seq1, qual1)
    trimmer = UmiTrimmer(number_of_bases=5)
    trimmed_read1 = trimmer(read1)
    assert (
        trimmed_read1.sequence
        == "AGCTTTTGTAGACAAGTGCTGTGCAGCTGATGTCAAAGAGACCTGCTTTGCTCTGGAGGGTCCAAAACTTGTAGCCTCAACCCGAGAAGCCATAGCCTAA"
    )
    assert trimmed_read1.umi == "TTTGC"
    read2 = Sequence("read2", seq2, qual2)
    trimmed_read2 = trimmer(read2)
    assert (
        trimmed_read2.sequence
        == "CTATGGCTTCTCGAGTTGAAGCTACAAGTTTTGGACCCTCCAGAGCAAAGCAGGTCTCTTTGACATCAGCTGCACAGCACTTGTCTACAAAAGCTGCAAAAGATCGGAAGAGCGTCTCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGACGTATCATTAAAAAAAAAAACACATCACATCAACAAGATAACACGACTTCTCCATCCACAGTACCGATGACCTCAACATTAGT"
    )
    assert trimmed_read2.umi == "ATAGG"
    umi_to_name = SyncUmi(delim="_")
    new_read1, new_read2 = umi_to_name(trimmed_read1, trimmed_read2)
    assert new_read1.name == "read1_TTTGC_ATAGG"
    assert new_read2.name == "read2_TTTGC_ATAGG"

    # 5 bases from read1, 0 from read2
    read1 = Sequence("read1", seq1, qual1)
    trimmer = UmiTrimmer(number_of_bases=5)
    trimmed_read1 = trimmer(read1)
    read2 = Sequence("read2", seq2, qual2)
    trimmer = UmiTrimmer(number_of_bases=0)
    untrim_read2 = trimmer(read2)
    assert (
        untrim_read2.sequence
        == "ATAGGCTATGGCTTCTCGAGTTGAAGCTACAAGTTTTGGACCCTCCAGAGCAAAGCAGGTCTCTTTGACATCAGCTGCACAGCACTTGTCTACAAAAGCTGCAAAAGATCGGAAGAGCGTCTCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGACGTATCATTAAAAAAAAAAACACATCACATCAACAAGATAACACGACTTCTCCATCCACAGTACCGATGACCTCAACATTAGT"
    )
    assert untrim_read2.umi is None
    umi_to_name = SyncUmi(delim="_")
    new_read1, new_read2 = umi_to_name(trimmed_read1, untrim_read2)
    assert new_read1.name == "read1_TTTGC"
    assert new_read2.name == "read2_TTTGC"

    # 0 bases from read1, 5 from read2
    read1 = Sequence("read1", seq1, qual1)
    trimmer = UmiTrimmer(number_of_bases=0)
    untrim_read1 = trimmer(read1)
    read2 = Sequence("read2", seq2, qual2)
    trimmer = UmiTrimmer(number_of_bases=5)
    trimmed_read2 = trimmer(read2)
    umi_to_name = SyncUmi(delim="_")
    new_read1, new_read2 = umi_to_name(untrim_read1, trimmed_read2)
    assert new_read1.name == "read1_ATAGG"
    assert new_read2.name == "read2_ATAGG"


def test_umi_trim_single():
    seq1 = "TTTGCAGCTTTTGTAGACAAGTGCTGTGCAGCTGATGTCAAAGAGACCTGCTTTGCTCTGGAGGGTCCAAAACTTGTAGCCTCAACCCGAGAAGCCATAGCCTAA"
    qual1 = "CCCCCFCGGGGGBFFAFC<?BEADCCF<FFFFGFFDFDFFGGGGCFGGC?DFFFEC;,===??DG==DDDFFFFG8DDD7+5;;DF*=)))10885D**58>6=0"
    # 5 bases UMI from read1
    read1 = Sequence("read1", seq1, qual1)
    trimmer = UmiTrimmer(number_of_bases=5)
    trimmed_read1 = trimmer(read1)
    assert (
        trimmed_read1.sequence
        == "AGCTTTTGTAGACAAGTGCTGTGCAGCTGATGTCAAAGAGACCTGCTTTGCTCTGGAGGGTCCAAAACTTGTAGCCTCAACCCGAGAAGCCATAGCCTAA"
    )
    assert trimmed_read1.umi == "TTTGC"
    add_umi = AddUmi(delim="_")
    assert add_umi(trimmed_read1).name == "read1_TTTGC"
    read1 = Sequence("read1", seq1, qual1)
    trimmer = UmiTrimmer(number_of_bases=0)
    trimmed_read1 = trimmer(read1)
    assert add_umi(trimmed_read1).umi is None
    assert add_umi(trimmed_read1).name == "read1"



# coding: utf-8
from atropos.commands.trim.qualtrim import twocolor_trim_index
from atropos.io.sequence import Sequence


def test_twocolor_trim():
    s = Sequence("n", "", "")
    assert twocolor_trim_index(s, cutoff=22) == 0
    s = Sequence(
        "n",
        "TCTCGTATGCCGTCTTATGCTTGAAAAAAAAAAGGGGGGGGGGGGGGGGGNNNNNNNNNNNGGNGG",
        "AA//EAEE//A6///E//A//EA/EEEEEEAEA//EEEEEEEEEEEEEEE###########EE#EA",
    )
    assert twocolor_trim_index(s, cutoff=22) == 33



import functools
import math
from typing import Iterable, Optional, Sequence, Tuple

from atropos.utils.collections import CountingDict


class RandomMatchProbability:
    """
    Class for computing random match probability for DNA sequences based on binomial
    expectation. Maintains a cache of factorials to speed computation.

    Args:
        init_size: Initial cache size.
    """

    def __init__(self, init_size: int = 150):
        self.cache = {}
        self.factorials = [1] * init_size
        self.max_n = 1
        self.cur_array_size = init_size

    def __call__(
        self,
        matches: int,
        size: int,
        match_prob: float = 0.25,
        mismatch_prob: float = 0.75,
    ) -> float:
        """
        Computes the random-match probability for a given sequence size and number of
        matches.

        Args:
            matches: Number of matches (numerator)
            size: Size of the sequence (denominator)
            match_prob: Probability of two random bases matching.
            mismatch_prob: Probability of two random bases not matcing.

        Returns:
            The probability.
        """
        # First see if we have the result in the cache
        key = (matches, size, match_prob)
        prob = self.cache.get(key, None)
        if prob:
            return prob

        # When there are no mismatches, the probability is
        # just that of observing a specific sequence of the
        # given length by chance.
        if matches == size:
            prob = match_prob ** matches
        else:
            nfac = self.factorial(size)
            prob = 0.0
            for i in range(matches, size + 1):
                j = size - i
                # use integer division in the case that the numbers are too
                # large for floating point division
                try:
                    div = nfac / self.factorial(i) / self.factorial(j)
                except OverflowError:
                    div = nfac // self.factorial(i) // self.factorial(j)
                prob += (mismatch_prob ** j) * (match_prob ** i) * div

        self.cache[key] = prob

        return prob

    def factorial(self, num: int) -> int:
        """
        Returns `num`!.
        """
        if num > self.max_n:
            self._fill_upto(num)
        return self.factorials[num]

    def _fill_upto(self, num: int) -> None:
        if num >= self.cur_array_size:
            extension_size = num - self.cur_array_size + 1
            self.factorials += [1] * extension_size

        idx = self.max_n
        next_i = idx + 1

        while idx < num:
            self.factorials[next_i] = next_i * self.factorials[idx]
            idx = next_i
            next_i += 1

        self.max_n = idx


class Histogram(CountingDict):
    """
    CountingDict that returns a summary dict that contains summary stats.
    """

    def summarize(self) -> dict:
        hist = super().summarize()
        return dict(hist=hist, summary=self.get_summary_stats())

    def get_summary_stats(self) -> dict:
        """Returns dict with mean, median, and modes of histogram.
        """
        values = tuple(self.keys())
        counts = tuple(self.values())
        mu0 = weighted_mean(values, counts)
        return dict(
            mean=mu0,
            stdev=weighted_stdev(values, counts, mu0),
            median=weighted_median(values, counts),
            modes=weighted_modes(values, counts),
        )


def mean(values: Sequence[float]) -> float:
    """Computes the mean of a sequence of numeric values.

    Args:
        values: Sequence of numeric values.

    Returns:
        The mean (floating point).
    """
    if len(values) == 0:
        raise ValueError("Cannot determine the mode of an empty sequence")
    return sum(values) / len(values)


def weighted_mean(values: Sequence[float], counts: Sequence[int]) -> float:
    """
    Computes the mean of a sequence of numeric values weighted by counts.

    Args:
        values: Sequence of numeric values.
        counts: Sequence of counts.

    Returns:
        The weighted mean (floating point).
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the mena of an empty sequence")
    if datalen != len(counts):
        raise ValueError("'values' and 'counts' must be the same length")
    return sum(v * c for v, c in zip(values, counts)) / sum(counts)


def stdev(values: Sequence[float], mu0: Optional[float] = None) -> float:
    """
    Computes the standard deviation of values having the specified mean.

    Args:
        values: The values from which to compute the standard deviation.
        mu0: The population mean. Computed from `values` if None.

    Returns:
        The standard deviation (floating point). Note: if `values` is of length 1,
        typically a ValueError would be raised, but this function returns 0.

    Raises:
        ValueError: If `values` is empty.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the stdev of an empty sequence")
    if datalen == 1:
        return 0
    if mu0 is None:
        mu0 = mean(values)
    return math.sqrt(sum((val - mu0) ** 2 for val in values) / (len(values) - 1))


def weighted_stdev(
    values: Sequence[float], counts: Sequence[int], mu0: Optional[float] = None
) -> float:
    """
    Computes the standard deviation of values having the specified mean weighted
    by counts.

    Args:
        values: The values from which to compute the standard deviation.
        counts: Sequence of counts.
        mu0: The population mean. Computed from `values` if None.

    Returns:
        The standard deviation (floating point). Note: if `values` is of length 1,
        typically a ValueError would be raised, but this function returns 0.

    Raises:
        ValueError: If `values` is empty or if `values` and `counts` are different
        lengths.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the stdev of an empty sequence")
    if datalen != len(counts):
        raise ValueError("'values' and 'counts' must be the same length")
    if datalen == 1:
        return 0
    if mu0 is None:
        mu0 = weighted_mean(values, counts)
    return math.sqrt(
        sum(((val - mu0) ** 2) * count for val, count in zip(values, counts))
        / (sum(counts) - 1)
    )


def median(values: Sequence[float]) -> float:
    """
    Median function borrowed from python statistics module, and sped up by in-place
    sorting of the array.

    Args:
        values: Sequence of numeric values.

    Returns:
        The median (floating point).

    Raises:
        ValueError if `values` is empty.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the median of an empty sequence")

    values = sorted(values)
    idx = datalen // 2

    if datalen % 2 == 1:
        return values[idx]
    else:
        return (values[idx - 1] + values[idx]) / 2


def weighted_median(values: Sequence[float], counts: Sequence[int]) -> Optional[float]:
    """
    Computes the median of `values` weighted by `counts`.

    Args:
        values: Sequence of unique values.
        counts: Sequence of counts, where each count is the number of times the
            value at the corresponding position appears in the sample.

    Returns:
        The weighted median, or None if the total of `counts` is 0.

    Raises:
        ValueError if `values` is empty or a different length than `counts`.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the median of an empty sequence")
    if datalen != len(counts):
        raise ValueError("'values' and 'counts' must be the same length")

    counts_cumsum = functools.reduce(lambda c, x: c + [c[-1] + x], counts, [0])[1:]
    total = counts_cumsum[-1]
    if total == 0:
        return None

    mid1 = mid2 = (total // 2) + 1
    if total % 2 == 0:
        mid1 -= 1
    val1 = val2 = None
    for i, val in enumerate(counts_cumsum):
        if val1 is None and mid1 <= val:
            val1 = values[i]
        if mid2 <= val:
            val2 = values[i]
            break

    return float(val1 + val2) / 2


def modes(values: Sequence[float]) -> Sequence[float]:
    """
    Computes a sorted sequence of the modal (i.e. most frequent) values.

    Args:
        values: The values for which to find the mode(s).

    Returns:
        A sequence of the modal value(s).

    Raises:
        ValueError if `values` is empty.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the mode of an empty sequence")
    elif datalen == 1:
        return values
    return _find_modes(CountingDict(values).items())


def weighted_modes(values: Sequence[float], counts: Sequence[int]) -> Sequence[float]:
    """
    Computes a sorted sequence of the modal (i.e. most frequent) values weighted by
    counts.

    Args:
        values: Sequence of unique values.
        counts: Sequence of counts, where each count is the number of times the
            value at the corresponding position appears in the sample.

    Returns:
        A sequence of the modal value(s).

    Raises:
        ValueError if `values` is empty or a different length than `counts`.
    """
    datalen = len(values)
    if datalen == 0:
        raise ValueError("Cannot determine the mode of an empty sequence")
    if datalen != len(counts):
        raise ValueError("'values' and 'counts' must be the same length")
    if datalen == 1:
        return values
    return _find_modes(zip(values, counts))


def _find_modes(value_count_iter: Iterable[Tuple[float, int]]) -> Sequence[float]:
    sorted_counts = sorted(value_count_iter, key=lambda x: x[1], reverse=True)
    modal_values = [sorted_counts[0][0]]
    mode_count = sorted_counts[0][1]

    for value, count in sorted_counts[1:]:
        if count == mode_count:
            modal_values.append(value)
        else:
            break

    modal_values.sort()

    return modal_values



"""
Atropos version {}

usage: atropos [--config <config file>] <command> [options]

commands
--------
{}

optional arguments:
  -h, --help                show this help message and exit
  --config <config file>    provide options in a config file

Use "atropos <command> --help" to see all options for a specific command.
See http://atropos.readthedocs.org/ for full documentation.

Atropos is a fork of Cutadapt 1.10 (
https://github.com/marcelm/cutadapt/tree/2f3cc0717aa9ff1e0326ea6bcb36b712950d4999)
by John Didion, et al., "Atropos: sensitive, specific, and speedy trimming of
NGS reads, submitted.

Cutadapt (https://github.com/marcelm/cutadapt) was developed by Marcel Martin,
"Cutadapt Removes Adapter Sequences From High-Throughput Sequencing Reads,"
EMBnet Journal, 2011, 17(1):10-12.
"""
import sys
from typing import Dict, Optional, Sequence, TypeVar, cast

from loguru import logger
import pkg_resources
from xphyle import open_

from atropos import __version__
from atropos.commands.console import CommandConsole
from atropos.utils import ReturnCode
from atropos.utils.paths import as_readable_path


HELP_TEXT = __doc__
T = TypeVar("T")


def run_atropos(args: Optional[Sequence[str]] = None) -> ReturnCode:
    """
    Run atropos.

    Args:
        args: Command-line arguments. If None, `sys.argv[1:]` is used.

    Returns:
        The return code.
    """
    if check_importability():
        if args is None:
            args = sys.argv[1:]
        return execute_cli(args)
    else:
        logger.error(
            f"""
ERROR: A required extension module could not be imported because it is
incompatible with your system. A quick fix is to recompile the extension
modules with the following command:

{sys.executable} setup.py build_ext -i

See the documentation for alternative ways of installing the program.

The original error message follows.
"""
        )
        return ReturnCode.ERROR


def check_importability() -> bool:  # pragma: no cover
    """
    Check that Cython modules have been compiled.

    Returns:
        True if a cython-compiled module can be imported, otherwise False.
    """
    try:
        import atropos.aligner  # pylint: disable=unused-variable
        return True
    except ImportError as err:
        if "undefined symbol" in str(err):
            return False
        else:
            raise


def execute_cli(args: Sequence[str] = ()) -> ReturnCode:
    """
    Executes the Atropos command-line interface.

    The first argument is expected to be the command name. If not (i.e. args is
    empty or the first argument starts with a '-'), the 'trim' command is
    assumed. If the first argument is '-h' or '--help', the command-level help
    is printed.

    Args:
        args: Command-line arguments.

    Returns:
        The return code.
    """
    commands = discover_commands()

    if len(args) == 0 or args[0] in ("-h", "--help"):
        print(
            HELP_TEXT.format(
                __version__,
                "\n".join(cmd.get_help() for cmd in commands.values())
            ),
            file=sys.stderr
        )
        return ReturnCode.ERROR

    if args[0] == "--config":
        config_path = as_readable_path(args[1])
        args = args[2:]

        with open_(config_path, "rt") as config_file:
            config_args = list(
                token for line in config_file for token in line.rstrip().split()
            )
    else:
        config_args = None

    def parse_command(_args):
        if not _args or _args[0][0] == "-":
            return "trim", _args
        else:
            return _args[0], _args[1:]

    if len(args) == 0:
        command_name, args = parse_command(config_args)
    else:
        command_name, args = parse_command(args)
        if config_args:
            args = config_args + args

    try:
        command = commands[command_name]
        retcode, summary = command.execute(args)
        if "exception" in summary:
            logger.opt(exception=summary["exception"]["details"]).error(
                f"Error executing command {command_name}"
            )
        return retcode
    except:
        logger.exception(f"Error executing command: {command_name}")
        return ReturnCode.ERROR


def discover_commands() -> Dict[str, CommandConsole]:
    """
    Discover Commands defined as entrypoints.
    """
    return dict(
        (cmd_class.name, cmd_class)
        for cmd_class in (
            cast(CommandConsole, entry_point.load(True))
            for entry_point in pkg_resources.iter_entry_points(group="atropos.commands")
        )
    )



import io

from loguru import logger
from pytest import raises

from atropos.commands.multicore import *


class TimeoutException(Exception):
    pass


log_capture_string = None


def setup():
    # Setup the console handler with a StringIO object
    global log_capture_string
    log_capture_string = io.StringIO()
    logger.add(sink=log_capture_string, level="DEBUG")


def assert_log(expected):
    log_contents = log_capture_string.getvalue()
    log_capture_string.close()
    assert log_contents == expected


def test_wait_on():
    class callbacks:
        def __init__(self):
            self.i = 0
            self.j = 0

        def condition(self):
            self.i += 1
            return self.i >= 5

        def fail_callback(self):
            self.j += 1

    c = callbacks()
    wait_on(c.condition, wait_message="waiting", fail_callback=c.fail_callback)
    assert c.i == 5
    assert c.j == 4


# assert_log("DEBUG: waiting for {} seconds".format())
def test_timeout():
    with raises(TimeoutException):

        def condition():
            return False

        wait_on(condition, timeout=2, wait=1, timeout_callback=TimeoutException)


def test_enqueue_dequeue():
    q = Queue(1)
    enqueue(q, 1)
    assert dequeue(q) == 1


def test_enqueue_timeout():
    with raises(TimeoutException):
        q = Queue(1)
        q.put(1)
        enqueue(q, 2, timeout=1, block_timeout=2, timeout_callback=TimeoutException)


def test_dequeue_timeout():
    with raises(TimeoutException):
        dequeue(Queue(1), timeout=1, block_timeout=2, timeout_callback=TimeoutException)


# TODO: port tests from testparallel here
# Test worker vs writer compression
# Test without writer process



from abc import ABCMeta, abstractmethod
from collections import namedtuple
import csv
import math
from pathlib import Path
import sys
from typing import (
    Callable, Iterator, IO, Optional, Sequence as SequenceType, Tuple, Type,
    Union, cast
)

from xphyle import STDIN, xopen

from atropos.errors import FormatError, UnknownFileTypeError
from atropos.io import InputRead, guess_format_from_name
from atropos.io.sequence import Sequence, ColorspaceSequence
from atropos.utils import classproperty, truncate_string
from atropos.utils.ngs import ALPHABETS, Alphabet

from ._base import (
    SequenceReader,
    SequenceReaderBase,
    PrefetchSequenceReader,
    estimate_num_records,
    sequence_names_match
)
from ._readers import FastqReader


class FastaReader(PrefetchSequenceReader):
    """
    Reader for FASTA files.

    Args:
        path: A path or a file-like object. In both cases, the file may
            be compressed (.gz, .bz2, .xz).
        keep_linebreaks: Whether to keep newline characters in the sequence.
        sequence_factory: The class to use when creating new sequence objects.
    """

    @classproperty
    def file_format(cls) -> str:
        return "FASTA"

    @classproperty
    def delivers_qualities(cls) -> bool:
        return False

    @classproperty
    def has_qualfile(cls) -> bool:
        return False

    @classproperty
    def colorspace(cls) -> bool:
        return False

    @classproperty
    def interleaved(cls) -> bool:
        return False

    def __init__(
        self,
        path: Union[str, Path, Iterator[str]],
        keep_linebreaks: bool = False,
        **kwargs
    ):
        self._delimiter = "\n" if keep_linebreaks else ""
        super().__init__(path, **kwargs)

    def _iter(self):
        """
        Reads the next entry from the file (single entry at a time).
        """
        name = None
        seq = []

        for i, line in enumerate(self._file):
            # strip() also removes DOS line breaks
            line = line.strip()

            if not line:
                continue

            if line and line[0] == ">":
                if name is not None:
                    yield self._sequence_factory(
                        name=name,
                        sequence=self._delimiter.join(seq),
                        alphabet=self._alphabet
                    )

                name = line[1:]
                seq = []
            elif line and line[0] == "#":
                continue
            elif name is not None:
                seq.append(line)
            else:
                raise FormatError(
                    f"At line {i + 1}: Expected '>' at beginning of FASTA record, "
                    f"but got {truncate_string(line)!r}."
                )

        if name is not None:
            yield self._sequence_factory(name, self._delimiter.join(seq), None)

    def estimate_num_records(self):
        # TODO: this will underestimate the record size (and thus overestimate the
        #  total number of records) for FASTA files with wrapped sequence lines.
        record_size = sum(seq.size_in_bytes() for seq in self._first_seq)
        return estimate_num_records([self.name], record_size, 2, 1)


class NgstreamSequenceReader(SequenceReader):
    @classproperty
    def file_format(cls) -> str:
        return "FASTQ"

    @classproperty
    def delivers_qualities(cls) -> bool:
        return True

    @classproperty
    def has_qualfile(cls) -> bool:
        return False

    @classproperty
    def colorspace(cls) -> bool:
        return False

    @classproperty
    def interleaved(cls) -> bool:
        return False

    def __init__(self, reader, **kwargs):
        import ngstream

        protocol = cast(ngstream.Protocol, reader)

        self._input_read = InputRead.PAIRED if protocol.paired else InputRead.SINGLE
        super().__init__(protocol, **kwargs)

    @property
    def input_read(self) -> InputRead:
        return self._input_read

    def estimate_num_records(self):
        return self._file.read_count

    def __iter__(self):
        if self._input_read == InputRead.PAIRED:
            for read in self._file:
                yield tuple(self._as_sequence(read[i]) for i in (0, 1))
        else:
            for read in self._file:
                yield self._as_sequence(read),

    def _as_sequence(self, frag):
        return self._sequence_factory(
            frag.name,
            frag.sequence,
            frag.qualities,
            alphabet=self._alphabet
        )

    def close(self):
        self._file.finish()


class FastaQualReader(SequenceReaderBase):
    """
    Reader for reads that are stored in .(CS)FASTA and .QUAL files.
    """

    @classproperty
    def file_format(cls) -> str:
        return "FastaQual"

    @classproperty
    def delivers_qualities(cls) -> bool:
        return True

    @classproperty
    def has_qualfile(cls) -> bool:
        return True

    @classproperty
    def colorspace(cls) -> bool:
        return False

    @classproperty
    def interleaved(cls) -> bool:
        return False

    @classproperty
    def input_read(self) -> InputRead:
        return InputRead.SINGLE

    def __init__(
        self,
        fastafile: Union[str, Path, Iterator[str]],
        qualfile: Union[str, Path, Iterator[str]],
        quality_base: int = 33,
        alphabet: Optional[Alphabet] = None,
        sequence_factory: Callable[..., Sequence] = Sequence,
    ):
        """
        Args:
            fastafile: filename or file-like object.
            qualfile: filename or file-like object.
            quality_base:
            sequence_factory: The class to use when creating new sequence objects.
            alphabet:
        """
        self._fastareader = FastaReader(fastafile)
        self._qualreader = FastaReader(qualfile, keep_linebreaks=True)
        self._quality_base = quality_base
        self._alphabet = alphabet
        self._sequence_factory = sequence_factory

    @property
    def input_names(self) -> Tuple[Tuple[str, str], None]:
        return (self._fastareader.name, self._qualreader.name), None

    @property
    def quality_base(self) -> Optional[int]:
        return self._quality_base

    def estimate_num_records(self):
        return self._fastareader.estimate_num_records()

    def __iter__(self):
        """
        Yields Sequence objects.
        """
        # conversion dictionary: maps strings to the appropriate ASCII-encoded
        # character
        conv = dict((str(i), chr(i + 33)) for i in range(-5, 256 - 33))

        for fastaread, qualread in zip(self._fastareader, self._qualreader):
            if fastaread.name != qualread.name:
                raise FormatError(
                    f"The read names in the FASTA and QUAL file do not match "
                    f"({fastaread.name!r} != {qualread.name!r})"
                )

            try:
                qualities = "".join(
                    [conv[value] for value in qualread.sequence.split()]
                )
            except KeyError as err:
                raise FormatError(
                    f"Within read named {fastaread.name!r}: Found invalid quality "
                    f"value {err}"
                )

            yield self._sequence_factory(
                name=fastaread.name,
                sequence=fastaread.sequence,
                qualities=qualities,
                alphabet=self._alphabet
            )

    def close(self):
        """
        Closes the underlying files.
        """
        self._fastareader.close()
        self._qualreader.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class ColorspaceSequenceReaderMixin:
    """
    Base class for colorspace sequence readers.
    """

    @classproperty
    def colorspace(cls) -> bool:
        return True

    def __init__(
        self,
        reader,
        quality_base: int = 33,
        sequence_factory: Callable[..., Sequence] = ColorspaceSequence,
        **kwargs
    ):
        super().__init__(
            reader,
            quality_base=quality_base,
            sequence_factory=sequence_factory,
            **kwargs
        )


class ColorspaceNgstreamSequenceReader(
    ColorspaceSequenceReaderMixin, NgstreamSequenceReader
):
    """
    Reads colorspace sequences from an SRA accession.
    """


class ColorspaceFastaReader(ColorspaceSequenceReaderMixin, FastaReader):
    """
    Reads colorspace sequences from a FASTA.
    """


class ColorspaceFastqReader(ColorspaceSequenceReaderMixin, FastqReader):
    """
    Reads colorspace sequences from a FASTQ.
    """


class ColorspaceFastaQualReader(ColorspaceSequenceReaderMixin, FastaQualReader):
    """
    Reads sequences and qualities from separate files and returns
    :class:`ColorspaceSequence`s.
    """


class PairedSequenceReader(SequenceReaderBase):
    """
    Reads paired-end reads from two files. Wraps two SequenceReader instances,
    making sure that reads are properly paired.
    """

    @classproperty
    def has_qualfile(cls) -> bool:
        return False

    @classproperty
    def interleaved(cls) -> bool:
        return False

    @classproperty
    def input_read(self) -> InputRead:
        return InputRead.PAIRED

    def __init__(
        self,
        file1,
        file2,
        quality_base: int = 33,
        colorspace: bool = False,
        file_format: bool = None,
        alphabet: Optional[Alphabet] = None
    ):
        """
        Args:
            file1, file2: The pair of files.
            colorspace: Whether the sequences are in colorspace.
            file_format: A file_format instance.
        """
        self._reader1 = open_reader(
            file1,
            colorspace=colorspace,
            quality_base=quality_base,
            file_format=file_format,
            alphabet=alphabet,
        )
        self._reader2 = open_reader(
            file2,
            colorspace=colorspace,
            quality_base=quality_base,
            file_format=file_format,
            alphabet=alphabet,
        )

    @property
    def input_names(self):
        return self._reader1.input_names[0], self._reader2.input_names[0]

    @property
    def file_format(self) -> str:
        return self._reader1.file_format

    def colorspace(self) -> bool:
        return self._reader1.colorspace

    @property
    def quality_base(self) -> int:
        return self._reader1.quality_base

    def delivers_qualities(self) -> bool:
        return self._reader1.delivers_qualities

    def estimate_num_records(self) -> Optional[int]:
        ests = list(
            filter(
                None,
                (
                    self._reader1.estimate_num_records(),
                    self._reader2.estimate_num_records()
                ),
            )
        )

        if len(ests) > 0:
            return max(ests)

    def __getattr__(self, name):
        return getattr(self._reader1, name)

    def __iter__(self):
        """
        Iterates over the paired reads. Each item is a pair of Sequence objects.
        """
        # Avoid usage of zip() below since it will consume one item too many.
        it1 = iter(self._reader1)
        it2 = iter(self._reader2)

        while True:
            try:
                read1 = next(it1)
            except StopIteration:
                # End of file 1. Make sure that file 2 is also at end.
                try:
                    next(it2)

                    raise FormatError(
                        "Reads are improperly paired. There are more reads in file 2 "
                        "than in file 1."
                    )
                except StopIteration:
                    pass

                break

            try:
                read2 = next(it2)
            except StopIteration:
                raise FormatError(
                    "Reads are improperly paired. There are more reads in file 1 than "
                    "in file 2."
                )

            if not sequence_names_match(read1, read2):
                raise FormatError(
                    f"Reads are improperly paired. Read name '{read1.name}' in file 1 "
                    f"does not match '{read2.name}' in file 2."
                )

            yield read1, read2

    def close(self):
        """
        Closes the underlying files.
        """
        self._reader1.close()
        self._reader2.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class SequenceReaderWrapper(SequenceReaderBase):
    def __init__(self, reader: SequenceReaderBase):
        """
        Args:
            reader: The reader to wrap.
        """
        self._reader = reader

    @property
    def file_format(self) -> str:
        return self._reader.file_format

    @property
    def delivers_qualities(self) -> bool:
        return self._reader.delivers_qualities

    @property
    def colorspace(self) -> bool:
        return self._reader.colorspace

    @property
    def input_names(self) -> Tuple[Union[str, SequenceType[str]], Optional[str]]:
        return self._reader.input_names

    @property
    def quality_base(self) -> int:
        return self._reader.quality_base

    @property
    def has_qualfile(self) -> bool:
        return self._reader.has_qualfile

    @property
    def interleaved(self) -> bool:
        return self._reader.interleaved

    @property
    def input_read(self) -> InputRead:
        return self._reader.input_read

    def estimate_num_records(self) -> Optional[int]:
        return self._reader.estimate_num_records()

    def __iter__(self):
        return iter(self._reader)

    def close(self):
        self._reader.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()


class InterleavedSequenceReader(SequenceReaderWrapper):
    """
    Reads paired-end reads from an interleaved FASTQ file.
    """

    @classproperty
    def has_qualfile(cls) -> bool:
        return False

    @classproperty
    def interleaved(cls) -> bool:
        return True

    @classproperty
    def input_read(self) -> InputRead:
        return InputRead.PAIRED

    def __init__(
        self,
        path,
        quality_base: int = 33,
        colorspace: bool = False,
        file_format: Optional[str] = None,
        alphabet: Optional[Alphabet] = None
    ):
        """
        Args:
            path: The interleaved FASTQ file.
            colorspace: Whether the sequences are in colorspace.
            file_format: A file_format instance.
        """
        super().__init__(open_reader(
            path,
            quality_base=quality_base,
            colorspace=colorspace,
            file_format=file_format,
            alphabet=alphabet,
        ))

    def estimate_num_records(self) -> int:
        return math.ceil(self._reader.estimate_num_records() / 2)

    def __iter__(self) -> Iterator[Tuple[Sequence, Sequence]]:
        # Avoid usage of zip() below since it will consume one item too many.
        itr = iter(self._reader)

        for read1 in itr:
            try:
                read2 = next(itr)
            except StopIteration:
                raise FormatError(
                    "Interleaved input file incomplete: Last record has no " "partner."
                )

            if not sequence_names_match(read1, read2):
                raise FormatError(
                    f"Reads are improperly paired. Name {read1.name!r} (first) does "
                    f"not match {read2.name!r} (second)."
                )

            yield read1, read2


class PairedToSingleEndReader(SequenceReaderWrapper):
    """
    Wrapper that yields either first or second reads from a paired-end reader.
    """

    def __init__(self, reader: SequenceReaderBase, input_read: InputRead):
        super().__init__(reader)
        self._input_read = input_read

    def __iter__(self):
        idx = 1 if self._input_read == InputRead.READ2 else 0
        for record in self._reader:
            yield record[idx]


class SAMReader(SequenceReaderBase, metaclass=ABCMeta):
    """
    Reader for SAM/BAM files. Paired-end files must be name-sorted. Does not support
    secondary/supplementary reads.

    TODO: either drop support for BAM files (require user to pipe output from
     `samtools view`) and implement pure-python SAM parsing, or replace pysam with a
     different BAM-parsing library (pysam is not Python 3.8 compatible).
    """

    @classproperty
    def file_format(cls) -> str:
        return "SAM"

    @classproperty
    def delivers_qualities(cls) -> bool:
        return True

    @classproperty
    def has_qualfile(cls) -> bool:
        return False

    @classproperty
    def colorspace(cls) -> bool:
        return False

    def __init__(
        self,
        path: Union[str, Path, Iterator[str], Iterator[bytes]],
        quality_base: int = 33,
        sequence_factory: Type[Sequence] = Sequence,
        alphabet: Alphabet = None,
        pysam_kwargs: Optional[dict] = None,
    ):
        """
        Args:
            path: A filename or a file-like object. If a filename, then .gz files
                are supported.
            sequence_factory: The class to use when creating new sequence objects.
        """
        self._close_on_exit = is_path = isinstance(path, (str, Path))
        self._quality_base = quality_base
        self._sequence_factory = sequence_factory
        self._alphabet = alphabet

        if is_path:
            self.name = str(path)
        elif hasattr(path, "name"):
            self.name = getattr(path, "name")
        else:
            self.name = path.__class__

        if self.name.endswith(".sam"):
            if is_path:
                self._file = xopen(path, "rt")
            else:
                self._file = cast(IO, path)

            self._sam_iter = SAMParser(self._file)
        elif self.name.endswith(".bam"):
            if is_path:
                self._file = xopen(path, "rb")
            else:
                self._file = cast(IO, path)

            if pysam_kwargs is None:
                pysam_kwargs = {"check_sq": False}

            self._sam_iter = BAMParser(self._file, **pysam_kwargs)
        else:
            # Todo: open file and check for BAM magic number
            raise ValueError(f"Cannot detect type of file {path}")

    @property
    def input_names(self):
        return self.name, None

    @property
    def quality_base(self) -> int:
        return self._quality_base

    def estimate_num_records(self):
        return self._sam_iter.estimate_num_records()

    def __iter__(self):
        # pysam raises an error of the SAM/BAM header is not complete,
        # even though it's unnecessary for our puroses. In that case,
        # we use our own simple SAM parser.
        return self._iter(self._sam_iter)

    @abstractmethod
    def _iter(self, sam):
        """
        Creates an iterator over records in the SAM/BAM file.
        """

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """
        Closes the underling AlignmentFile.
        """
        if self._close_on_exit and self._file is not None:
            self._file.close()
            self._file = None

    def _as_sequence(self, read):
        return self._sequence_factory(
            read.query_name,
            read.query_sequence,
            read.query_qualities,
            alphabet=self._alphabet
        )


SAMRead = namedtuple(
    "SAMRead",
    ("query_name", "query_sequence", "query_qualities", "is_read1", "is_read2"),
)


class SAMParser:
    def __init__(self, sam_file: IO):
        self._sam_file = sam_file
        self._reader = csv.reader(sam_file, delimiter="\t")
        self._header_size = 0

        line = None

        for line in self._reader:
            # skip header lines
            if not line[0].startswith("@"):
                self._header_size += len(line)
                break

        self._next_line = line

    def estimate_num_records(self):
        record_len = len("\t".join(self._next_line))
        return estimate_num_records(
            self._sam_file, record_len, 1, header_size=self._header_size
        )

    def __iter__(self):
        return self

    def __next__(self):
        if self._next_line is None:
            raise StopIteration()

        is_read1 = (int(self._next_line[1]) & 64) > 0

        read = SAMRead(
            self._next_line[0],
            self._next_line[9],
            self._next_line[10],
            is_read1,
            not is_read1,
        )

        try:
            self._next_line = next(self._reader)
        except StopIteration:
            self._next_line = None

        return read


class BAMParser:
    def __init__(self, bam_file: IO, **kwargs):
        import pysam

        self._reader = pysam.AlignmentFile(str(bam_file), **kwargs)

    def estimate_num_records(self):
        """
        Todo: Not sure how to estimate the number of records in a name-sorted BAM file.
         There is no index, and bgzip doesn't have an option to get the uncompressed
         size. Not sure if gzip -l will work correctly for a bgzip file.
        """

    def __iter__(self):
        return self

    def __next__(self):
        read = next(self._reader)

        return SAMRead(
            read.query_name,
            read.query_sequence,
            "".join(chr(33 + q) for q in read.query_qualities),
            read.is_read1,
            read.is_read2,
        )


class SingleEndSAMReader(SAMReader):
    """
    Reader for single-end SAM/BAM files.
    """
    @classproperty
    def interleaved(cls) -> bool:
        return False

    @property
    def input_read(self) -> InputRead:
        return InputRead.SINGLE

    def _iter(self, sam):
        for read in sam:
            yield self._as_sequence(read)


class PairedEndEstimatorMixin:
    def estimate_num_records(self):
        est = cast(SequenceReader, super()).estimate_num_records()
        if est is None:
            return None
        else:
            return math.ceil(est / 2)


class Read1SingleEndSAMReader(SAMReader, PairedEndEstimatorMixin):
    """
    Reads a paired-end SAM/BAM file as if it were single-end, yielding only the
    first read from each pair.
    """
    @classproperty
    def interleaved(cls) -> bool:
        return False

    @property
    def input_read(self) -> InputRead:
        return InputRead.READ1

    def _iter(self, sam):
        for read in sam:
            if read.is_read1:
                yield self._as_sequence(read)


class Read2SingleEndSAMReader(SAMReader, PairedEndEstimatorMixin):
    """
    Reads a paired-end SAM/BAM file as if it were single-end, yielding only the
    second read from each pair.
    """

    @classproperty
    def interleaved(cls) -> bool:
        return False

    @property
    def input_read(self) -> InputRead:
        return InputRead.READ2

    def _iter(self, sam):
        for read in sam:
            if read.is_read2:
                yield self._as_sequence(read)


class PairedEndSAMReader(SAMReader, PairedEndEstimatorMixin):
    """
    Reads pairs of reads from a SAM/BAM file. The file must be name-sorted.
    """

    @classproperty
    def interleaved(cls) -> bool:
        return True

    @property
    def input_read(self) -> InputRead:
        return InputRead.PAIRED

    def _iter(self, sam):
        for reads in zip(sam, sam):
            if reads[0].query_name != reads[1].query_name:
                raise FormatError(
                    f"Consecutive reads {reads[0].query_name}, {reads[1].query_name} "
                    f"in paired-end SAM/BAM file do not have the same name; make sure "
                    f"your file is name-sorted and does not contain any "
                    f"secondary/supplementary alignments."
                )

            if not reads[0].is_read1 ^ reads[1].is_read1:
                raise ValueError(
                    f"There must be exactly one read with read1 flag set in pair "
                    f"{reads}"
                )

            if not reads[0].is_read2 ^ reads[1].is_read2:
                raise ValueError(
                    f"There must be exactly one read with read2 flag set in pair "
                    f"{reads}"
                )

            if reads[1].is_read1:
                # Swap reads so the first read is always read1
                reads = (reads[1], reads[0])

            yield tuple(self._as_sequence(r) for r in reads)


class FileWithPrependedLine:
    """
    A file-like object that allows to "prepend" a single line to an already
    opened file. That is, further reads on the file will return the provided
    line and only then the actual content. This is needed to solve the problem
    of autodetecting input from a stream: As soon as the first line has been
    read, we know the file type, but also that line is "gone" and unavailable
    for further processing.

    Args:
        file: An already opened file-like object.
        line: A single string (newline will be appended if not included).
    """

    def __init__(self, file: IO, line: str):
        if not line.endswith("\n"):
            line += "\n"
        self.first_line = line
        self._file = file

    @property
    def name(self) -> str:
        return self._file.name

    def __iter__(self):
        yield self.first_line
        yield from self._file

    def close(self):
        """
        Closes the underlying file.
        """
        self._file.close()


def open_reader(
    file1=None,
    file2=None,
    ngstream_reader=None,
    qualfile: Union[str, Path, Iterator[str]] = None,
    quality_base: Optional[int] = None,
    colorspace: bool = False,
    file_format: Optional[str] = None,
    interleaved: bool = False,
    input_read: Optional[InputRead] = None,
    alphabet: Optional[Alphabet] = None,
) -> SequenceReaderBase:
    """
    Open sequence files in FASTA or FASTQ format for reading. This is a factory that
    returns an instance of one of the ...Reader classes also defined in this module.

    Args:
        file1: Path to read1 regular or compressed file or file-like object.
        file2: Path to read2 regular or compressed file or file-like object.
        ngstream_reader: An ngstream.Protocol object.
        qualfile: Path to qualfile regular or compressed file or file-like object. If
            specified, then file1 must be a FASTA file and sequences are single-end.
            One of file2 and qualfile must always be None (no paired-end data is
            supported when reading qualfiles).
        quality_base: Base for quality values.
        colorspace: If True, instances of the Colorspace... classes
            are returned.
        file_format: If set to None, file format is autodetected from the file
            name extension. Set to 'fasta', 'fastq', 'sra-fastq', 'sam', or
            'bam' to not auto-detect. Colorspace is not auto-detected and must
            always be requested explicitly.
        interleaved: If True, then file1 contains interleaved paired-end data.
            file2 and qualfile must be None in this case.
        input_read: When file1 is a paired-end interleaved or SAM/BAM
            file, this specifies whether to only use the first or second read
            (1 or 2) or to use both reads (3).
        alphabet: An Alphabet instance - the alphabet to use to validate sequences.
    """
    if interleaved and (file2 is not None or qualfile is not None):
        raise ValueError("When interleaved is set, file2 and qualfile must be None")

    if file2 is not None and qualfile is not None:
        raise ValueError("Setting both file2 and qualfile is not supported")

    if alphabet and isinstance(alphabet, str):
        if alphabet not in ALPHABETS:
            raise ValueError(f"Invalid alphabet {alphabet}")

        alphabet = ALPHABETS[alphabet]

    if file2 is not None:
        return PairedSequenceReader(
            file1,
            file2,
            quality_base=quality_base,
            colorspace=colorspace,
            file_format=file_format,
            alphabet=alphabet,
        )

    if qualfile is not None:
        # read from .(CS)FASTA/.QUAL
        reader_class = ColorspaceFastaQualReader if colorspace else FastaQualReader
        return reader_class(
            file1,
            qualfile=qualfile,
            quality_base=quality_base,
            alphabet=alphabet
        )

    reader = None

    if file1:
        if file_format is None:
            if file1 not in (STDIN, sys.stdin):
                file_format = guess_format_from_name(file1)

            if file_format is None:
                if file1 == STDIN:
                    file1 = sys.stdin

                for line in file1:
                    if line.startswith("#"):
                        # Skip comment lines (needed for csfasta)
                        continue

                    if line.startswith(">"):
                        file_format = "fasta"
                    elif line.startswith("@"):
                        file_format = "fastq"

                    # TODO: guess SAM/BAM from data
                    file1 = FileWithPrependedLine(file1, line)

                    break

        if file_format is not None:
            file_format = file_format.lower()

            if file_format in ("sam", "bam"):
                if colorspace:
                    raise ValueError(
                        "SAM/BAM format is not currently supported for colorspace reads"
                    )

                if interleaved:
                    return PairedEndSAMReader(
                        file1, quality_base=quality_base, alphabet=alphabet
                    )
                elif input_read == InputRead.READ1:
                    return Read1SingleEndSAMReader(
                        file1, quality_base=quality_base, alphabet=alphabet
                    )
                elif input_read == InputRead.READ2:
                    return Read2SingleEndSAMReader(
                        file1, quality_base=quality_base, alphabet=alphabet
                    )
                else:
                    return SingleEndSAMReader(
                        file1, quality_base=quality_base, alphabet=alphabet
                    )

            if interleaved:
                reader = InterleavedSequenceReader(
                    file1,
                    quality_base=quality_base,
                    colorspace=colorspace,
                    file_format=file_format,
                    alphabet=alphabet
                )
            elif file_format == "fasta":
                reader_class = ColorspaceFastaReader if colorspace else FastaReader
                reader = reader_class(file1, alphabet=alphabet)
            elif file_format == "fastq":
                reader_class = ColorspaceFastqReader if colorspace else FastqReader
                reader = reader_class(
                    file1, quality_base=quality_base, alphabet=alphabet
                )
            elif file_format == "sra-fastq" and colorspace:
                def sra_colorspace_sequence_factory(
                    name: str,
                    sequence: str,
                    qualities: str,
                    *args,
                    **kwargs
                ):
                    """
                    Factory for an SRA colorspace sequence (which has one quality
                    value too many).
                    """
                    return ColorspaceSequence(
                        name, sequence, qualities[1:], *args, **kwargs
                    )

                reader = ColorspaceFastqReader(
                    file1,
                    quality_base=quality_base,
                    alphabet=alphabet,
                    sequence_factory=sra_colorspace_sequence_factory
                )
    elif ngstream_reader:
        if colorspace:
            reader_class = ColorspaceNgstreamSequenceReader
        else:
            reader_class = NgstreamSequenceReader

        reader = reader_class(
            ngstream_reader,
            quality_base=quality_base,
            alphabet=alphabet
        )

    if reader:
        if input_read == InputRead.PAIRED or not reader.paired:
            return reader
        else:
            return PairedToSingleEndReader(reader, input_read)

    raise UnknownFileTypeError(
        f"File format {file_format!r} is unknown (expected 'sra-fastq' (only for "
        f"colorspace), 'fasta', 'fastq', 'sam', or 'bam')."
    )



from typing import Optional

from atropos.errors import FormatError
from atropos.utils import truncate_string
from atropos.utils.ngs import Alphabet

from ._sequence import Sequence


class ColorspaceSequence(Sequence):
    """
    Sequence object for colorspace reads.
    """

    def __init__(
        self,
        name: str,
        sequence: str,
        qualities: Optional[str] = None,
        primer: Optional[str] = None,
        name2: str = "",
        original_length: Optional[int] = None,
        match=None,
        match_info=None,
        clipped=None,
        insert_overlap: bool = False,
        merged: bool = False,
        corrected: int = 0,
        umi: Optional[str] = None,
        alphabet: Optional[Alphabet] = None,
    ):
        # In colorspace, the first character is the last nucleotide of the
        # primer base and the second character encodes the transition from the
        # primer base to the first real base of the read.
        if primer is None:
            self.primer = sequence[0:1]
            sequence = sequence[1:]
        else:
            self.primer = primer

        if self.primer not in ("A", "C", "G", "T"):
            raise FormatError(
                f"Primer base is {self.primer!r} in read {truncate_string(name)!r}, "
                f"but it should be one of A, C, G, T."
            )

        if qualities is not None and len(sequence) != len(qualities):
            rname = truncate_string(name)
            raise FormatError(
                f"In read named {rname!r}: length of colorspace quality sequence "
                f"({len(qualities)}) and length of read ({len(sequence)}) do not match "
                f"(primer is: {self.primer!r})"
            )

        super().__init__(
            name,
            sequence,
            qualities,
            name2,
            original_length,
            match,
            match_info,
            clipped,
            insert_overlap,
            merged,
            corrected,
            umi,
            alphabet,
        )

    def __repr__(self):
        fmt_str = "<ColorspaceSequence(name={0!r}, primer={1!r}, sequence={2!r}{3})>"

        if self.qualities is None:
            qstr = ""
        else:
            qstr = f", qualities={truncate_string(self.qualities)!r}"

        return fmt_str.format(
            truncate_string(self.name),
            self.primer,
            truncate_string(self.sequence),
            qstr,
        )

    def __getitem__(self, key):
        return self.__class__(
            self.name,
            self.sequence[key],
            self.qualities[key] if self.qualities is not None else None,
            self.primer,
            self.name2,
            self.original_length,
            self.match,
            self.match_info,
            self.clipped,
            self.insert_overlap,
            self.merged,
            self.corrected,
            self.umi,
        )



"""
Build atropos.

Cython is run when
* no pre-generated C sources are found,
* or the pre-generated C sources are out of date,
* or when --cython is given on the command line.
"""
from pathlib import Path
import sys

from setuptools import setup, Extension, find_packages
from distutils.version import LooseVersion
from distutils.command.sdist import sdist
from distutils.command.build_ext import build_ext

version_info = sys.version_info
if sys.version_info < (3, 6):
    sys.stdout.write("At least Python 3.6 is required.\n")
    sys.exit(1)


MIN_CYTHON_VERSIONS = {
    (3, 6): "0.25.2",
    (3, 7): "0.29",
    (3, 8): "0.29.14"
}
min_cython_version = MIN_CYTHON_VERSIONS[version_info[0:2]]


class BuildExtCython(build_ext):
    def run(self):
        # If we encounter a PKG-INFO file, then this is likely a .tar.gz/.zip
        # file retrieved from PyPI that already includes the pre-cythonized
        # extension modules, and then we do not need to run cythonize().
        if (Path.cwd() / "PKG-INFO").exists():
            no_cythonize(extensions)
        else:
            # Otherwise, this is a 'developer copy' of the code, and then the
            # only sensible thing is to require Cython to be installed.
            check_cython_version()
            from Cython.Build import cythonize
            self.extensions = cythonize(self.extensions)

        super().run()


class SdistCython(sdist):
    def run(self):
        # Make sure the compiled Cython files in the distribution are up-to-date
        from Cython.Build import cythonize
        check_cython_version()
        cythonize(extensions)
        super().run()


def out_of_date(_extensions):
    """
    Check whether any pyx source is newer than the corresponding generated
    C source or whether any C source is missing.
    """
    for extension in _extensions:
        for pyx in extension.sources:
            pyx_path = Path(pyx)
            if pyx_path.suffix not in (".pyx", ".py"):
                continue
            if extension.language == "c++":
                csource = pyx_path.with_suffix(".cpp")
            else:
                csource = pyx_path.with_suffix(".c")
            # When comparing modification times, allow five seconds slack:
            # If the installation is being run from pip, modification
            # times are not preserved and therefore depends on the order in
            # which files were unpacked.
            if not csource.exists() or (
                pyx_path.stat().st_mtime > csource.stat().st_mtime + 5
            ):
                return True

    return False


def check_cython_version():
    """Exit if Cython was not found or is too old"""
    try:
        from Cython import __version__ as cyversion
    except ImportError:
        sys.stdout.write(
            f"ERROR: Cython is not installed. Install at least Cython version "
            f"{min_cython_version} to continue.\n"
        )
        sys.exit(1)
    if LooseVersion(cyversion) < LooseVersion(min_cython_version):
        sys.stdout.write(
            f"ERROR: Your Cython is at version '{cyversion}', but at least "
            f"version {min_cython_version} is required.\n"
        )
        sys.exit(1)


def no_cythonize(_extensions, **_ignore):
    """
    Change file extensions from .pyx to .c or .cpp.

    Copied from Cython documentation
    """
    for extension in _extensions:
        sources = []
        for sfile in extension.sources:
            sfile_path = Path(sfile)
            if sfile_path.suffix in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = sfile_path.with_suffix(ext)
            sources.append(sfile)
        extension.sources[:] = list(str(src) for src in sources)


with open(
    Path(__file__).parent.absolute() / "README.md", encoding="utf-8"
) as f:
    readme = f.read()

# Define extensions to be Cythonized
extensions = [
    Extension("atropos.aligner._aligner", sources=["atropos/aligner/_aligner.pyx"]),
    Extension(
        "atropos.commands.trim._qualtrim",
        sources=["atropos/commands/trim/_qualtrim.pyx"],
    ),
    Extension(
        "atropos.io.sequence._sequence",
        sources=["atropos/io/sequence/_sequence.pyx"]
    ),
    Extension(
        "atropos.io.readers._readers",
        sources=["atropos/io/readers/_readers.pyx"]
    ),
]

install_requirements = [
    f"Cython>={min_cython_version}",
    "loguru>=0.4.0",
    "pokrok>=0.2.0",
    "xphyle>=4.2.1",
]
test_requirements = [
    "pytest",  # 'jinja2', 'pysam',
]
extra_requirements = {
    "progressbar": ["progressbar2"],
    "tqdm": ["tqdm"],
    "khmer": ["khmer"],
    "pysam": ["pysam"],
    "jinja": ["jinja2"],
    "sra": ["ngstream>=0.2.2"],
}

setup(
    name="atropos",
    use_scm_version=True,
    cmdclass={
        "build_ext": BuildExtCython,
        "sdist": SdistCython
    },
    author="John Didion",
    author_email="github@didion.net",
    url="https://atropos.readthedocs.org/",
    description="Specific, sensitive, and speedy trimming of NGS reads.",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="MIT",
    ext_modules=extensions,
    packages=find_packages(),
    package_data={
        "atropos": ["adapters/*.fa", "commands/**/templates/*"]
    },
    setup_requires=["setuptools_scm"],
    install_requires=install_requirements,
    tests_require=test_requirements,
    extras_require=extra_requirements,
    entry_points={
        "console_scripts": [
            "atropos=atropos.__main__:main"
        ],
        "atropos.commands": [
            "detect = atropos.commands.detect.console:DetectCommandConsole",
            "error = atropos.commands.error.console:ErrorCommandConsole",
            "qc = atropos.commands.qc.console:QcCommandConsole",
            "trim = atropos.commands.trim.console:TrimCommandConsole",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: MIT License",
        "License :: Public Domain",
        "Natural Language :: English",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        #"Programming Language :: Python :: 3.8",
    ],
)



# kate: syntax Python;
# cython: profile=False, emit_code_comments=False
import copy

from atropos.errors import FormatError
from atropos.utils import truncate_string
from atropos.utils.ngs import reverse_complement


cdef class Sequence(object):
    """
    A record in a FASTQ file. Also used for FASTA (then the qualities attribute
    is None). qualities is a string and it contains the qualities encoded as
    ascii(qual+33).

    If an adapter has been matched to the sequence, the 'match' attribute is
    set to the corresponding Match instance.

    Todo:
        `name` should be `description` (reserve .name for the part before the first
        space)
    """
    cdef:
        public str name
        public str sequence
        public str qualities
        public str name2
        public int original_length
        public object match
        public object match_info
        public object clipped
        public bint insert_overlap
        public bint merged
        public int corrected
        public str umi

    def __init__(
        self, str name, str sequence, str qualities=None, str name2='',
        original_length=None, match=None, match_info=None, clipped=None,
        insert_overlap=False, merged=False, corrected=0, str umi=None, alphabet=None,
    ):
        # Validate sequence and qualities lengths are equal
        if qualities is not None:
            slen = len(sequence)
            qlen = len(qualities)

            # check that sequence and qualities are the same length
            if slen != qlen:
                rname = truncate_string(name)
                raise FormatError(
                    f"In read named {rname!r}: length of quality sequence ({qlen}) and "
                    f"length  of read ({slen}) do not match"
                )

        # If an alphabet is specified, replace any bases not in the alphabet
        # with the default character.
        if alphabet:
            sequence = alphabet.resolve_string(sequence)

        self.name = name
        self.sequence = sequence
        self.qualities = qualities
        self.name2 = name2
        self.original_length = original_length or len(sequence)
        self.match = match
        self.match_info = match_info
        self.clipped = clipped or [0,0,0,0]
        self.insert_overlap = insert_overlap
        self.merged = merged
        self.corrected = corrected
        self.umi = umi

    def subseq(self, begin=0, end=None):
        if end is None:
            new_read = self[begin:]
        else:
            new_read = self[begin:end]
        end_bases = len(self) - end
        offset = 2 if self.match else 0
        if begin:
            new_read.clipped[offset] += begin
        if end_bases:
            new_read.clipped[offset+1] += end_bases
        return begin, end_bases, new_read

    def clip(self, front=0, back=0):
        if back < 0:
            new_read = self[front:back]
            back *= -1
        else:
            new_read = self[front:]
        offset = 2 if self.match else 0
        if front:
            new_read.clipped[offset] += front
        if back:
            new_read.clipped[offset+1] += back
        return front, back, new_read

    def reverse_complement(self):
        """
        Returns a copy of this read with the sequence reverse-complemented
        and the qualities reversed.

        TODO: add option to also reverse the alignment positions
        """
        sequence = reverse_complement(self.sequence)
        qualities = clipped = match_info = None
        if self.qualities:
            qualities = ''.join(reversed(self.qualities))
        if self.match_info:
            match_info = [copy.copy(m) for m in self.match_info]

        new_read = self.__class__(
            self.name,
            sequence,
            qualities,
            self.name2,
            self.original_length,
            None,
            match_info,
            list(self.clipped),
            self.insert_overlap,
            self.merged,
            self.corrected,
            self.umi
        )

        if self.match:
            match = self.match.copy()
            match.read = new_read
            new_read.match = match

        return new_read

    @property
    def size_in_bytes(self) -> int:
        size = len(self.name) + len(self.sequence)
        if self.name2:
            size += len(self.name2)
        if self.qualities:
            size += len(self.qualities)
        return size

    def __getitem__(self, key):
        """slicing"""
        return self.__class__(
            self.name,
            self.sequence[key],
            self.qualities[key] if self.qualities is not None else None,
            self.name2,
            self.original_length,
            self.match,
            self.match_info,
            list(self.clipped),
            self.insert_overlap,
            self.merged,
            self.corrected,
            self.umi
        )

    def __repr__(self):
        qstr = ""
        if self.qualities is not None:
            qstr = f", qualities={truncate_string(self.qualities)!r}"
        return f"<Sequence(name={truncate_string(self.name)!r}, " \
               f"sequence={truncate_string(self.sequence)!r}{qstr})>"

    def __len__(self):
        return len(self.sequence)

    def __richcmp__(self, other, int op):
        if 2 <= op <= 3:
            eq = self.name == other.name and \
                self.sequence == other.sequence and \
                self.qualities == other.qualities
            if op == 2:
                return eq
            else:
                return not eq
        else:
            raise NotImplementedError()

    def __reduce__(self):
        return Sequence, (self.name, self.sequence, self.qualities, self.name2)
